# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound

from terms.core import exceptions
from terms.core import patterns
from terms.core.terms import Term, Predicate, isa, are, Time


class Lexicon(object):

    def __init__(self, session, config):
        self.config = config
        self.session = session
        self.word = self.get_term('word')
        self.verb = self.get_term('verb')
        self.noun = self.get_term('noun')
        self.number = self.get_term('number')
        self.exist = self.get_term('exist')
        self.endure = self.get_term('endure')
        self.exclusive_endure = self.get_term('exclusive-endure')
        self.occur = self.get_term('occur')
        self.happen = self.get_term('happen')
        self.vtime = self.get_term('time')
        self.thing = self.get_term('thing')
        self.finish = self.get_term('finish')
        self.time = self.session.query(Time).one()
        self.now_term = self.make_term(str(0 + self.time.now), self.number)
        self._term_cache = {}

    @classmethod
    def initialize(cls, session):
        '''
        Create basic terms.
        '''
        word = Term('word', _bootstrap=True)
        session.add(word)
        session.commit()
        word.term_type = word
        session.commit()
        verb = Term('verb', ttype=word, bases=(word,))
        session.add(verb)
        noun = Term('noun', ttype=word, bases=(word,))
        session.add(noun)
        number = Term('number', ttype=word, bases=(word,))
        session.add(number)
        exist = Term('exist', ttype=verb, bases=(word,),
                      objs={'subj': word})
        session.add(exist)
        endure = Term('endure', ttype=verb, bases=(exist,),
                       objs={'since_': number, 'till_': number})
        session.add(endure)
        exclusive_endure = Term('exclusive-endure', ttype=verb, bases=(endure,))
        session.add(exclusive_endure)
        occur = Term('occur', ttype=verb, bases=(exist,),
                   objs={'at_': number})
        session.add(occur)
        happen = Term('happen', ttype=verb, bases=(occur,))
        session.add(happen)
        time = Term('time', ttype=verb, bases=(exist,),
                    objs={'subj': number})
        session.add(time)
        thing = Term('thing', ttype=noun, bases=(word,))
        session.add(thing)
        finish = Term('finish', ttype=verb, bases=(occur,),
                      objs={'subj': thing, 'what': exist})
        session.add(finish)
        time = Time()
        session.add(time)
        session.commit()

    def get_term(self, name):
        '''
        Given a name (string), get a Term from the database.
        The Term must exist.
        '''
        cache = getattr(self, '_term_cache', None)
        if cache is not None:
            if name in cache:
                return cache[name]
        else:
            self._term_cache = {}
        try:
            return self.session.query(Term).filter_by(name=name).one()
        except MultipleResultsFound:
            raise exceptions.TermRepeated(name)
        except NoResultFound:
            raise exceptions.TermNotFound(name)

    def get_terms(self, term_type):
        '''
        Get all terms of type term_type.
        The term_type must exist.
        '''
        types = self.get_subterms(term_type)
        type_ids = [t.id for t in types if t is not self.number]
        return self.session.query(Term).filter(Term.type_id.in_(type_ids)).all()

    def make_term(self, name, term_type, **objs):
        '''
        Make a Term from a name (string) and a term_type (Term).
        Can also produce a predicate.
        The term is not saved or added to the session.
        '''
        try:
            return self.get_term(name)
        except exceptions.TermNotFound:
            if are(term_type, self.noun):
                return self._make_noun(name, ntype=term_type)
            elif are(term_type, self.thing):
                return self._make_name(name, term_type)
            elif are(term_type, self.verb):
                return self._make_verb(name, vtype=term_type, objs=objs)
            elif are(term_type, self.exist):
                return self.make_pred(name, term_type, **objs)
            elif term_type == self.number:
                return self.make_number(name)
            else:
                return Term(name, ttype=term_type)

    def make_subterm(self, name, super_terms, **objs):
        '''
        Make a Term from a name (string) and bases (Term's).
        The bases are the supertypes of a type,
        and can be a tuple of terms or a single term.
        The term is not saved or added to the session.
        '''
        try:
            return self.get_term(name)
        except exceptions.TermNotFound:
            if isa(super_terms, self.word):
                super_terms = (super_terms,)
            term_base = super_terms[0]
            if are(term_base, self.noun):
                return self._make_subnoun(name, bases=super_terms)
            elif are(term_base, self.thing):
                return self._make_noun(name, bases=super_terms)
            elif are(term_base, self.verb):
                return self._make_subverb(name, bases=super_terms)
            elif are(term_base, self.exist):
                return self._make_verb(name, bases=super_terms, objs=objs)

    def add_term(self, name, term_type, **objs):
        term = self.make_term(name, term_type, **objs)
        self.session.add(term)
        term.purge_cache(self)
        self._term_cache[name] = term
        return term

    def add_subterm(self, name, super_terms, **objs):
        term = self.make_subterm(name, super_terms, **objs)
        self.session.add(term)
        term.purge_cache(self)
        self._term_cache[name] = term
        return term

    def get_subterms(self, term):
        cache = getattr(term, '_sub_cache', None)
        if cache is not None:
            return cache
        name = term.name
        m = patterns.varpat.match(name)
        if m:
            if m.group(2):
                term = self.get_term(m.group(1).lower())
            else:
                return ()
        subtypes = set([term])
        self._recurse_subterms(term, subtypes)
        subterms = tuple(subtypes)
        term._sub_cache = subterms
        return subterms

    def make_var(self, name):
        '''
        Make a term that represents a variable in a rule or query.
        It is not added to the session.
        Its name has the original trailing digits.

        '''
        try:
            return self.get_term(name)
        except exceptions.TermNotFound:
            pass
        m = patterns.varpat.match(name)
        if m.group(2):
            basename = m.group(1).lower()
            bases = self.get_term(basename)
            var = self.make_subterm(name, bases)
        else:
            tname = m.group(1).lower()
            if len(tname) == 1:
                tvar = self.number
            else:
                tvar = self.get_term(tname)
            if isa(tvar, self.verb) or tvar == self.number:
                var = Term(name, ttype=tvar)
            else:
                var = self.make_term(name, tvar)
        var.var = True
        self.session.add(var)
        return var

    def _recurse_subterms(self, term, subterms):
        for st in term.subwords:
            if st.var:
                continue
            subterms.add(st)
            self._recurse_subterms(st, subterms)

    def _make_noun(self, name, bases=None, ntype=None):
        if bases is None:
            bases = (self.thing,)
        elif isa(bases, self.word):
            bases = (bases,)
        if ntype is None:
            ntype = self.noun
        return Term(name, ttype=ntype, bases=tuple(bases))

    def _make_subnoun(self, name, bases=None):
        if bases is None:
            bases = (self.noun,)
        return Term(name, ttype=self.word, bases=tuple(bases))

    def _make_name(self, name, noun_):
        return Term(name, ttype=noun_)

    def _make_verb(self, name, bases=None, vtype=None, objs=None):
        if objs is None:
            objs = {}
        for l in objs:
            if '_' in l:
                raise exceptions.IllegalLabel(l)
        if vtype is None:
            vtype = bases[0].term_type
        return Term(name, ttype=vtype, bases=tuple(bases), objs=objs)

    def _make_subverb(self, name, bases=None):
        return Term(name, ttype=self.word, bases=tuple(bases))

    def make_number(self, num):
        num = str(0 + eval(str(num), {}, {}))
        number = Term(num, ttype=self.number)
        number.number = True
        return number

    def make_pred(self, true, verb_, **objs):
        return Predicate(true, verb_, **objs)



# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

import time
import functools

from sqlalchemy import Column, Sequence, Index
from sqlalchemy import ForeignKey, Integer, String, Boolean
from sqlalchemy.orm import relationship, backref, aliased
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from sqlalchemy.exc import InvalidRequestError

from terms.core import localdata
from terms.core.terms import isa, are, get_bases
from terms.core.terms import Base, Term, term_to_base, Predicate
from terms.core.lexicon import Lexicon
from terms.core.factset import FactSet
from terms.core import exceptions
from terms.core.utils import Match, merge_submatches


class Network(object):

    def __init__(self, session, config):
        self.session = session
        self.config = config
        self.activations = []
        self.root = self.session.query(RootNode).one()
        self.lexicon = Lexicon(session, config)
        self.present = FactSet('present', self.lexicon, config)
        self.past = FactSet('past', self.lexicon, config)
        self.pipe = None

    @classmethod
    def initialize(self, session):
        try:
            session.query(RootNode).one()
        except NoResultFound:
            root = RootNode()
            session.add(root)
            Lexicon.initialize(session)

    def passtime(self):
        past = eval(self.now, {}, {})
        now = 0
        if self.config['time'] == 'normal':
            now = past + 1
        elif self.config['time'] == 'real':
            now = int(time.time())

        q = self.lexicon.make_var('Occur1')
        topast = self.present.query_facts(q, {})
        for f in topast:
            for m in f.matches:
                self.session.delete(m)
            f.matches = []
            f.pred.add_object('at_', self.lexicon.now_term)
            self.present.add_object_to_fact(f, self.lexicon.now_term,
                                                    ('at_', '_term'))
            f.factset = 'past'

        self.now = now

    def _get_now(self):
        return str(self.lexicon.time.now)

    def _set_now(self, val):
        self.lexicon.time.now = eval(str(val), {}, {})
        now = str(0 + self.lexicon.time.now)
        self.lexicon.now_term = self.lexicon.make_term(now, self.lexicon.number)

    now = property(_get_now, _set_now)

    def get_paths(self, pred):
        '''
        build a path for each testable feature in term.
        Each path is a tuple of strings,
        and corresponds to a node in the primary network.
        '''
        paths = []
        verb_ = pred.term_type
        self._recurse_paths(verb_, pred, paths, (), first=True)
        return paths

    def _recurse_paths(self, verb_, pred, paths, path, first=False):
        paths.append(path + ('_verb',))
        if not isa(pred, self.lexicon.verb):  # not a verb var
            paths.append(path + ('_neg',))
        for obt in sorted(verb_.object_types, key=lambda x: x.label):
            if obt.label in ('till_', 'at_'):
                continue
            t = obt.obj_type
            if isa(t, self.lexicon.verb):
                if obt.label in pred.objects:
                    pred = pred.get_object(obt.label)
                    verb_ = pred.term_type
                    self._recurse_paths(verb_, pred, paths, path + (obt.label,))
                else:
                    paths.append(path + (obt.label, '_verb'))
            else:
                paths.append(path + (obt.label, '_term'))

    def _get_nclass(self, ntype):
        mapper = Node.__mapper__
        return mapper.base_mapper.polymorphic_map[ntype].class_

    def add_fact(self, pred):
        factset = self.present
        if isa(pred, self.lexicon.exclusive_endure):
            old_pred = Predicate(pred.true, pred.term_type)
            old_pred.add_object('subj', pred.get_object('subj'))
            for label in pred.objects:
                if label.startswith('u-'):
                    old_pred.add_object(label, pred.get_object(label))
            self.finish(old_pred)
        elif isa(pred, self.lexicon.finish):
            tofinish = pred.get_object('what')
            self.finish(tofinish)
        #neg = pred.copy()
        #neg.true = not neg.true
        #contradiction = factset.query(neg)
        #if contradiction:
        #    raise exceptions.Contradiction('we already have ' + str(neg))

        facts = factset.query_facts(pred, {})
        if facts.count() == 0:
            if isa(pred, self.lexicon.endure):
                pred.add_object('since_', self.lexicon.now_term)
            fact = factset.add_fact(pred)
            if isa(pred, self.lexicon.happen):
                if self.pipe is not None:
                    self.pipe.send_bytes(str(pred).encode('utf8'))
            if self.root.child_path:
                m = Match(pred)
                m.paths = self.get_paths(pred)
                m.fact = fact
                Node.dispatch(self.root, m, self)
            n = 0
            while self.activations:
                n += 1
                cmc = int(self.config['commit_many_consecuences'])
                if cmc and n % cmc == 0:
                    self.session.commit()
                match = self.activations.pop(0)
                Node.dispatch(self.root, match, self)
            return fact
        else:
            return facts.first()

    def finish(self, predicate):
        fs = self.present.query_facts(predicate, {})
        for f in fs:
            pred = f.pred
            if isa(pred, self.lexicon.endure):
                self.present.add_object_to_fact(f, self.lexicon.now_term, ('till_', '_term'))
                f.factset = 'past'
                for m in f.matches:
                    self.session.delete(m)
                f.matches = []

    def del_fact(self, pred):
        fact = self.present.query_facts(pred, {}).one()
        self.session.delete(fact)

    def add_rule(self, prems, conds, condcode, cons):
        rule = Rule()
        for n, pred in enumerate(prems):
            vars = {}
            paths = self.get_paths(pred)
            old_node = self.root
            for path in paths:
                old_node = self.get_or_create_node(old_node, pred, path, vars, rule)
            if old_node.terminal:
                pnode = old_node.terminal
            else:
                pnode = PremNode(old_node)
                old_node.terminal = pnode
            premise = Premise(pnode, n, pred)
            rule.prems.append(premise)
            for n, varname in vars.values():
                rule.pvars.append(PVarname(premise, n, varname))
        rule.conditions = conds
        rule.condcode = condcode
        for con in cons:
            if isinstance(con, Predicate):
                rule.consecuences.append(con)
            else:
                rule.vconsecuences.append(con)
        for prem in rule.prems:
            matches = self.present.query(prem.pred)
            for match in matches:
                try:
                    prem.node.matches.filter(PMatch.fact==match.fact).one()
                except NoResultFound:
                    m = PMatch(prem.node, match.fact)
                    for var, val in match.items():
                        numvar = prem.name_to_num(var)
                        m.pairs.append(MPair.make_pair(numvar, val))
                prem.dispatch(match, self)
        return rule


    def query(self, *q):
        submatches = []
        for pred in q:
            factset = self.present
            if set(pred.objects).intersection({'at_', 'till_'}):
                factset = self.past
            smatches = factset.query(pred)
            submatches.append(smatches)
        matches = merge_submatches(submatches)
        unique = []
        for m in matches:
            for m2 in unique:
                if m2 == m:
                    break
            else:
                unique.append(m)
        return unique

    def get_or_create_node(self, parent, term, path, vars, rule):
        ntype_name = path[-1]
        cls = self._get_nclass(ntype_name)
        value = cls.resolve(term, path)
        redundant_var = getattr(value, 'redundant_var', False)
        is_var = getattr(value, 'var', False)
        if redundant_var and not is_var:
            value = value.term_type
        name = getattr(value, 'name', '')
        pnum = 0
        if is_var:
            if name not in vars:
                pnum = len(vars) + 1
                vars[name] = (pnum, Varname(value, rule))
            else:
                pnum = vars[name][0]
        rnum = 0
        if redundant_var:
            rname = redundant_var.name
            if rname not in vars:
                rnum = len(vars) + 1
                vars[rname] = (rnum, Varname(redundant_var, rule))
            else:
                rnum = vars[rname][0]
        try:
            nodes = self.session.query(cls).filter(cls.parent_id==parent.id, cls.var==pnum)
            if redundant_var:
                nodes = nodes.filter(cls.redundant_var==rnum)
            node = nodes.filter(cls.value==value).one()
        except NoResultFound:
            #  build the node and append it
            node = cls(value)
            node.var = pnum
            if redundant_var:
                node.redundant_var = rnum
            parent.children.append(node)
            if not parent.child_path:
                parent.child_path = path
        return node


class Node(Base):
    '''
    An abstact node in the primary (or premises) network.
    It is extended by concrete node classes.
    '''
    __tablename__ = 'nodes'
    id = Column(Integer, Sequence('node_id_seq'), primary_key=True)
    child_path_str = Column(String)
    var = Column(Integer, default=0, index=True)
    redundant_var = Column(Integer, default=0, index=True)
    parent_id = Column(Integer, ForeignKey('nodes.id'), index=True)
    children = relationship('Node',
                         backref=backref('parent',
                                         uselist=False,
                                         remote_side=[id]),
                         primaryjoin="Node.id==Node.parent_id",
                         cascade='all,delete-orphan',
                         lazy='dynamic')

    ntype = Column(String)
    __mapper_args__ = {'polymorphic_on': ntype}

    def __init__(self, value):
        self.value = value

    def _get_path(self):
        try:
            return self._path
        except AttributeError:
            try:
                self._path = tuple(self.child_path_str.split('.'))
            except AttributeError:
                return ()
            return self._path

    def _set_path(self, path):
        self.child_path_str = '.'.join(path)
        self._path = path

    child_path = property(_get_path, _set_path)

    @classmethod
    def resolve(cls, pred, path):
        '''
        Get the value pointed at by path in w (a word).
        It can be a boolean (for neg nodes),
        a sting (for label nodes),
        a word, or some custom value for custom node types.
        '''
        raise NotImplementedError

    @classmethod
    def dispatch(cls, parent, match, network):
        if parent.child_path:
            path = parent.child_path
            ntype_name = path[-1]
            chcls = network._get_nclass(ntype_name)
            value = chcls.resolve(match.pred, path)
            children = chcls.get_children(parent, value, network)
            for ch in children:
                for child in ch:
                    new_match = match.copy()
                    if child.var:
                        val = None
                        if chcls is VerbNode and isa(child.value, network.lexicon.exist):
                            val = TermNode.resolve(match.pred, path)
                        else:
                            val = value
                        if child.var in match and match[child.var] != val:
                            continue
                        new_match[child.var] = val
                    if chcls is VerbNode and child.redundant_var:
                        new_match[child.redundant_var] = TermNode.resolve(match.pred, path)
                    chcls.dispatch(child, new_match, network)
        if parent.terminal:
            parent.terminal.dispatch(match, network)

    @classmethod
    def get_children(cls, parent, value, factset):
        '''
        Get the value pointed at by path in w (a word).
        It can be a boolean (for neg nodes),
        a sting (for label nodes),
        a word, or some custom value for custom node types.
        '''
        raise NotImplementedError


class RootNode(Node):
    '''
    A root node
    '''
    __mapper_args__ = {'polymorphic_identity': '_root'}
    __tablename__ = 'rootnodes'
    nid = Column(Integer, ForeignKey('nodes.id'), primary_key=True)

    def __init__(self):
        pass


class NegNode(Node):
    '''
    A node that tests whether a predicate is negated
    '''
    __mapper_args__ = {'polymorphic_identity': '_neg'}
    __tablename__ = 'negnodes'
    nid = Column(Integer, ForeignKey('nodes.id'), primary_key=True)

    value = Column(Boolean, index=True)

    @classmethod
    def resolve(cls, pred, path):
        for segment in path[:-1]:
            pred = pred.get_object(segment)
        try:
            return pred.true
        except AttributeError:
            # Predicate variable
            return True

    @classmethod
    def get_children(cls, parent, value, network):
        return [network.session.query(cls).filter(cls.parent_id==parent.id, (cls.value==value) | (cls.value==None))]


class TermNode(Node):
    '''
    '''
    __mapper_args__ = {'polymorphic_identity': '_term'}
    __tablename__ = 'termnodes'
    nid = Column(Integer, ForeignKey('nodes.id'), primary_key=True)

    term_id = Column(Integer, ForeignKey('terms.id'), index=True)
    value = relationship('Term', primaryjoin="Term.id==TermNode.term_id")

    @classmethod
    def resolve(cls, term, path):
        '''
        Get the value pointed at by path in w (a word).
        It can be a boolean (for neg nodes),
        a sting (for label nodes),
        a word, or some custom value for custom node types.
        '''
        try:
            for segment in path[:-1]:
                term = term.get_object(segment)
        except (AttributeError, KeyError):
            return None
        return term

    @classmethod
    def get_children(cls, parent, value, network):
        if isa(value, network.lexicon.exist):
            types = (value.term_type.term_type,) + get_bases(value.term_type.term_type)
            type_ids = [t.id for t in types]
            return network.session.query(cls).filter(cls.parent_id==parent.id, Node.var>0).join(Term, cls.term_id==Term.id).filter(Term.type_id.in_(type_ids)),
        children = network.session.query(cls).filter(cls.parent_id==parent.id, (cls.value==value) | (cls.value==None))
        vchildren = ()
        if value is not None:
            types = (value.term_type,) + get_bases(value.term_type)
            type_ids = [t.id for t in types]
            if type_ids:
                vchildren = network.session.query(cls).filter(cls.parent_id==parent.id).join(Term, cls.term_id==Term.id).filter(Term.var==True).filter(Term.type_id.in_(type_ids))
#         if not isa(value, network.lexicon.thing) and not isa(value, network.lexicon.number):
#             bases = (value,) + get_bases(value)
#             tbases = aliased(Term)
#             base_ids = (b.id for b in bases)
#             if base_ids and vchildren:
#                 vchildren = vchildren.join(term_to_base, Term.id==term_to_base.c.term_id).join(tbases, term_to_base.c.base_id==tbases.id).filter(tbases.id.in_(base_ids))  # XXX can get duplicates
        return children, vchildren


class VerbNode(Node):
    '''
    '''
    __mapper_args__ = {'polymorphic_identity': '_verb'}
    __tablename__ = 'verbnodes'
    nid = Column(Integer, ForeignKey('nodes.id'), primary_key=True)

    verb_id = Column(Integer, ForeignKey('terms.id'), index=True)
    value = relationship('Term', primaryjoin="Term.id==VerbNode.verb_id")

    @classmethod
    def resolve(cls, term, path):
        try:
            for segment in path[:-1]:
                term = term.get_object(segment)
            if term.var or getattr(term, 'redundant_var', False):
                return term
            return term.term_type
        except (AttributeError, KeyError):
            return None

    @classmethod
    def get_children(cls, parent, value, network):
        children = network.session.query(cls).filter(cls.parent_id==parent.id, (cls.value==value) | (cls.value==None))
        pchildren, vchildren = [], []
        if value is not None:
            types = (value,) + get_bases(value)
            type_ids = [t.id for t in types]
            chvars = network.session.query(cls).filter(cls.parent_id==parent.id, Node.var>0)
            pchildren = chvars.join(Term, cls.verb_id==Term.id).filter(Term.type_id.in_(type_ids))
            tbases = aliased(Term)
            vchildren = chvars.join(Term, cls.verb_id==Term.id).join(term_to_base, Term.id==term_to_base.c.term_id).join(tbases, term_to_base.c.base_id==tbases.id).filter(tbases.id.in_(type_ids))
        return children, pchildren, vchildren


class PremNode(Base):
    '''
    a terminal node for a premise
    '''
    __tablename__ = 'premnodes'

    id = Column(Integer, Sequence('premnode_id_seq'), primary_key=True)
    parent_id = Column(Integer, ForeignKey('nodes.id'), index=True)
    parent = relationship('Node', backref=backref('terminal', uselist=False,
                                                  cascade='all,delete-orphan'),
                         primaryjoin="Node.id==PremNode.parent_id")

    def __init__(self, parent):
        self.parent = parent  # node

    def dispatch(self, match, network):
        if not self.prems[0].check_match(match, network):
            return
        m = PMatch(self, match.fact)
        for var, val in match.items():
            m.pairs.append(MPair.make_pair(var, val))
        for premise in self.prems:
            nmatch = premise.num_to_names(match)
            premise.dispatch(nmatch, network)

from sqlalchemy.sql import select
from sqlalchemy.sql.expression import Select

class NoMatches(Exception):
    pass

class Premise(Base):
    '''
    Relation between rules and premnodes
    '''
    __tablename__ = 'premises'

    id = Column(Integer, Sequence('premise_id_seq'), primary_key=True)
    prem_id = Column(Integer, ForeignKey('premnodes.id'), index=True)
    node = relationship('PremNode', backref='prems',
                         primaryjoin="PremNode.id==Premise.prem_id")
    rule_id = Column(Integer, ForeignKey('rules.id'), index=True)
    rule = relationship('Rule', backref=backref('prems',
                                                 cascade='all,delete-orphan'))
                         # primaryjoin="Rule.id==Premise.rule_id")
    pred_id = Column(Integer, ForeignKey('predicates.id'), index=True)
    pred = relationship('Predicate',
                         primaryjoin="Predicate.id==Premise.pred_id")
    order = Column(Integer)

    def __init__(self, pnode, order, pred):
        self.node = pnode
        self.order = order
        self.pred = pred

    def check_match(self, match, network):
        pred = self.pred
        prem_paths = network.present.get_paths(pred)
        for p in prem_paths:
            if p[-1] == '_num':
                p = p[:-1] + ('_term',)
            if p not in match.paths:
                return False
        return True

    def num_to_names(self, match):
        nmatch = match.copy()
        for num, o in match.items():
            pvar = tuple(filter(lambda x: x.rule==self.rule and x.num==num, self.pvars))[0]
            name = pvar.varname.name
            nmatch[name] = o
            del nmatch[num]
        return nmatch

    def name_to_num(self, name):
        pvar = tuple(filter(lambda x: x.varname.name==name, self.pvars))[0]
        return pvar.num

    def dispatch(self, match, network):
        prems = [p for p in self.rule.prems if p != self]
        try:
            if prems:
                matches = self.recurse_premises(match, prems, network)
            else:
                matches = [match]
        except NoMatches:
            return
        for m in matches:
            self.rule.dispatch(m, network)

    def recurse_premises(self, match, remaining_prems, network):
        prem, pmatches = self.pick_prem(remaining_prems, match, network)
        remaining_prems.remove(prem)
        matches = []
        for pm in pmatches:
            new_match = match.copy()
            for mpair in pm.pairs:
                vname = self.rule.get_varname(prem, mpair.var)
                if vname not in new_match:
                    new_match[vname] = mpair.val
            try:
                passes = self.rule.test_conditions(new_match, network)
            except KeyError:
                passes = True
            if passes:
                matches.append(new_match)
        if not remaining_prems:
            return matches
        else:
            new_matches = []
            for m in matches:
                try:
                    new_matches += self.recurse_premises(m, remaining_prems[:], network)
                except NoMatches:
                    pass
            return new_matches

    def pick_prem(self, prems, match, network):
        count, pmatches, picked = float('inf'), None, None
        prems.sort(key=lambda p: p.node.matches.count())
        for prem in prems:
            pms = prem.filter_pmatches(match, network)
            newcount = pms.count() if pms else 0
            if newcount == 0:
                raise NoMatches
            elif newcount == 1:
                return prem, pms
            if newcount < count:
                count, pmatches, picked = newcount, pms, prem
        return picked, pmatches

    def filter_pmatches(self, match, network):
        pmatches = self.node.matches
        pvar_map = self.rule.get_pvar_map(match, self)
        subqueries = []
        for var, val in pvar_map:
            apair = aliased(MPair)
            if isinstance(val, Predicate):
                cpair = aliased(PPair.__table__)
                ccol = cpair.c.pred_id
            else:
                cpair = aliased(TPair.__table__)
                ccol = cpair.c.term_id
            subquery = select([apair.parent_id], from_obj=[apair, cpair], whereclause=(apair.var==var)&(ccol==val.id)&(apair.id==cpair.c.mid), distinct=True)
            subqueries.append(subquery)
        if subqueries:
            def count_subquery(q):
                n = network.session.execute(aliased(q).count()).scalar()
                if n == 0:
                    raise NoMatches
                return n
            subqueries.sort(key=count_subquery)
            subquery = functools.reduce(Select.intersect, subqueries)
            pmatches = pmatches.filter(PMatch.id.in_(subquery)).distinct(PMatch.id)
        return pmatches


class PMatch(Base):
    __tablename__ = 'pmatchs'

    id = Column(Integer, Sequence('pmatch_id_seq'), primary_key=True)
    prem_id = Column(Integer, ForeignKey('premnodes.id'), index=True)
    prem = relationship('PremNode', backref=backref('matches', lazy='dynamic',
                                                    cascade='all,delete-orphan'),
                         primaryjoin="PremNode.id==PMatch.prem_id")
    fact_id = Column(Integer, ForeignKey('facts.id'), index=True)
    fact = relationship('Fact', backref=backref('matches', cascade='all,delete-orphan'),
                         primaryjoin="Fact.id==PMatch.fact_id")

    def __init__(self, prem, fact):
        self.prem = prem
        self.fact = fact


class MPair(Base):
    __tablename__ = 'mpairs'

    id = Column(Integer, Sequence('mpair_id_seq'), primary_key=True)
    parent_id = Column(Integer, ForeignKey('pmatchs.id'), index=True)
    parent = relationship('PMatch', backref=backref('pairs',
                                                    cascade='all,delete-orphan'),
                         primaryjoin="PMatch.id==MPair.parent_id")
    var = Column(Integer, index=True)
    mindex = Index('mindex', 'id', 'parent_id')

    mtype = Column(Integer)
    __mapper_args__ = {'polymorphic_on': mtype}

    def __init__(self, var, val):
        self.var = var
        self.val = val

    @classmethod
    def make_pair(cls, var, val):
        if isinstance(val, Predicate):
            return PPair(var, val)
        else:
            return TPair(var, val)

class TPair(MPair):
    __tablename__ = 'tpairs'
    __mapper_args__ = {'polymorphic_identity': 0}
    mid = Column(Integer, ForeignKey('mpairs.id'), primary_key=True)
    term_id = Column(Integer, ForeignKey('terms.id'), index=True)
    val = relationship('Term', primaryjoin="Term.id==TPair.term_id")
    tindex = Index('tindex', 'mid', 'term_id')


class PPair(MPair):
    __tablename__ = 'ppairs'
    __mapper_args__ = {'polymorphic_identity': 1}
    mid = Column(Integer, ForeignKey('mpairs.id'), primary_key=True)
    pred_id = Column(Integer, ForeignKey('predicates.id'), index=True)
    val = relationship('Predicate',
                         primaryjoin="Predicate.id==PPair.pred_id")
    pindex = Index('pindex', 'mid', 'pred_id')


class PVarname(Base):
    """
    Mapping from varnames in rules (pvars belong in rules)
    to premise, number.
    Premises have numbered variables;
    and different rules can share a premise,
    but translate differently its numbrered vars to varnames.
    """
    __tablename__ = 'pvarnames'


    id = Column(Integer, Sequence('pvarname_id_seq'), primary_key=True)
    rule_id = Column(Integer, ForeignKey('rules.id'), index=True)
    rule = relationship('Rule', backref=backref('pvars', lazy='joined',
                                                cascade='all,delete-orphan'),
                         primaryjoin="Rule.id==PVarname.rule_id")
    prem_id = Column(Integer, ForeignKey('premises.id'), index=True)
    prem = relationship('Premise', backref=backref('pvars', lazy='joined'),
                         primaryjoin="Premise.id==PVarname.prem_id")
    varname_id = Column(Integer, ForeignKey('varnames.id'), index=True)
    varname = relationship('Varname', backref='pvarnames',
                         lazy='joined',
                         primaryjoin="Varname.id==PVarname.varname_id")
    num = Column(Integer, index=True)

    def __init__(self, prem, num, varname):
        self.prem = prem
        self.num = num
        self.varname = varname


class Varname(Base):
    """
    a variable in a rule,
    it has a name
    """
    __tablename__ = 'varnames'

    id = Column(Integer, Sequence('varname_id_seq'), primary_key=True)
    rule_id = Column(Integer, ForeignKey('rules.id'), index=True)
    rule = relationship('Rule', backref=backref('varnames', lazy='joined',
                                                cascade='all,delete-orphan'),
                         primaryjoin="Rule.id==Varname.rule_id")
    term_id = Column(Integer, ForeignKey('terms.id'), index=True)
    var = relationship('Term', backref='varnames',
                         lazy='joined',
                         primaryjoin="Term.id==Varname.term_id")

    def __init__(self, var, rule):
        self.var = var
        self.rule = rule

    def _get_name(self):
        return self.var.name

    name = property(_get_name)


class Rule(Base):
    '''
    '''
    __tablename__ = 'rules'

    id = Column(Integer, Sequence('rule_id_seq'), primary_key=True)

    def test_conditions(self, match, network):
        for cond in self.conditions:
            if not cond.test(match, network):
                return False
        return True

    def dispatch(self, match, network):

        if not self.test_conditions(match, network):
            return

        if self.condcode:
            if not self.condcode.test(match, network):
                return

        cons = []
        for con in self.consecuences:
            cons.append(con.substitute(match))

        for con in self.vconsecuences:
            cons.append(match[con.name])

        for con in cons:
            factset = network.present
            if isa(con, network.lexicon.exclusive_endure):
                old_pred = Predicate(con.true, con.term_type)
                old_pred.add_object('subj', con.get_object('subj'))
                for label in con.objects:
                    if label.startswith('u-'):
                        old_pred.add_object(label, con.get_object(label))
                network.finish(old_pred)
            elif isa(con, network.lexicon.finish):
                tofinish = con.get_object('what')
                network.finish(tofinish)
            # XXX make contradiction configurabe
            #neg = con.copy()
            #neg.true = not neg.true
            #contradiction = factset.query(neg)
            #if contradiction:
            #    raise exceptions.Contradiction('we already have ' + str(neg))
            if factset.query_facts(con, {}).count() == 0:
                if isa(con, network.lexicon.endure):
                    con.add_object('since_', network.lexicon.now_term)
                fact = factset.add_fact(con)
                if isa(con, network.lexicon.happen):
                    if network.pipe is not None:
                        network.pipe.send_bytes(str(con).encode('utf8'))
                if network.root.child_path:
                    m = Match(con)
                    m.paths = network.get_paths(con)
                    m.fact = fact
                    network.activations.append(m)

    def get_pvar_map(self, match, prem):
        pvar_map = []
        for name, val in match.items():
            pvars = filter(lambda x: x.prem==prem and x.varname.var.name==name, self.pvars)
            for n, pvar in enumerate(pvars):
                pvar_map.append((pvar.num, val))
                if n > 1:
                    raise exceptions.Corruption('should not happen')
        return pvar_map

    def get_varname(self, prem, num):
        pvar = tuple(filter(lambda x: x.prem==prem and x.num==num, self.pvars))[0]
        return pvar.varname.name


class CondArg(Base):
    '''
    '''
    __tablename__ = 'condargs'

    id = Column(Integer, Sequence('condarg_id_seq'), primary_key=True)
    cond_id = Column(Integer, ForeignKey('conditions.id'), index=True)
    cond = relationship('Condition', backref=backref('args',
                                                     cascade='all,delete-orphan'),
                         primaryjoin="Condition.id==CondArg.cond_id")
    term_id = Column(Integer, ForeignKey('terms.id'), index=True)
    term = relationship('Term',
                         primaryjoin="Term.id==CondArg.term_id")

    def __init__(self, term):
        self.term = term

    def solve(self, match):
        if self.term.var:
            return match[self.term.name]
        return self.term


class Condition(Base):
    '''
    '''
    __tablename__ = 'conditions'

    id = Column(Integer, Sequence('condition_id_seq'), primary_key=True)
    rule_id = Column(Integer, ForeignKey('rules.id'), index=True)
    rule = relationship('Rule', backref=backref('conditions',
                                                 cascade='all,delete-orphan'),
                         primaryjoin="Rule.id==Condition.rule_id")

    ctype = Column(Integer)
    __mapper_args__ = {'polymorphic_on': ctype}

    def __init__(self, *args):
        for arg in args:
            self.args.append(CondArg(arg))


class CondIsa(Condition):
    __mapper_args__ = {'polymorphic_identity': 0}

    def test(self, match, network):
        return isa(self.args[0].solve(match), self.args[1].solve(match))


class CondIs(Condition):
    __mapper_args__ = {'polymorphic_identity': 1}

    def test(self, match, network):
        return are(self.args[0].solve(match), self.args[1].solve(match))


class CondCode(Base):
    __tablename__ = 'condcodes'

    id = Column(Integer, Sequence('condcode_id_seq'), primary_key=True)
    rule_id = Column(Integer, ForeignKey('rules.id'), index=True)
    rule = relationship('Rule', backref=backref('condcode', uselist=False,
                                                 cascade='all,delete-orphan'),
                         primaryjoin="Rule.id==CondCode.rule_id")
    code = Column(String)

    def test(self, match, network):
        exec_locals = {'condition': True}
        exec_locals['match'] = match
        for k, v in match.items():
            if getattr(v, 'number', False):
                exec_locals[k] = eval(v.name, {}, {})
            else:
                exec_locals[k] = v
        try:
            exec(self.code, localdata.exec_globals, exec_locals)
        except Exception:
            if exec_locals['condition']:
                raise
            return False

        for k, v in exec_locals.items():
            if k in ('condition', '__builtins__', 'match'):
                continue
            try:
                match[k] = network.lexicon.make_term(str(0 + v), network.lexicon.number)
            except TypeError:
                pass
        return exec_locals['condition']

    def __init__(self, code):
        self.code = code



# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

from threading import local

localdata = local()

localdata.exec_globals = {'__builtins__': __builtins__}


def register_exec_global(fun, name=''):
    if name:
        localdata.exec_globals[name] = fun
    else:
        try:
            localdata.exec_globals[fun.__name__] = fun
        except AttributeError:  # method?
            localdata.exec_globals[fun.__func__.__name__] = fun


def count(compiler, sen):
    resp = compiler.parse(sen + '?')
    if resp == 'false':
        return 0
    elif resp == 'true':
        return 1
    return len(resp)

register_exec_global(count)




from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from terms.core.network import Network
from terms.core.terms import Base


def get_sasession(config):
    address = '%s/%s' % (config['dbms'], config['dbname'])
    engine = create_engine(address)
    Session = sessionmaker(bind=engine)
    if config['dbname'] == ':memory:':
        session = Session()
        Base.metadata.create_all(engine)
        Network.initialize(session)
        session.close()
    return Session




import sys
import os

try:
    import readline
    HAS_READLINE = True
except ImportError:
    HAS_READLINE = False

from code import InteractiveConsole

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from terms.core import register_exec_global
from terms.core.utils import get_config
from terms.core.compiler import Compiler, Runtime


class TermsRepl(object):

    def __init__(self, config):
        self.config = config
        self._buffer = ''
        self.no_response = object()
        self.prompt = '>> '
        if HAS_READLINE and config['terms_history_file'] and int(config['terms_history_length']):
            readline.set_history_length(int(config['terms_history_length']))
            fn = os.path.expanduser(config['terms_history_file'])
            try:
                if not os.path.exists(fn):
                    with open(fn, 'w') as f:
                        f.write('# terms history\n')
                readline.read_history_file(fn)
            except Exception:
                pass
        address = '%s/%s' % (config['dbms'], config['dbname'])
        engine = create_engine(address)
        Session = sessionmaker(bind=engine)
        if config['dbname'] == ':memory:':
            from terms.core.terms import Base
            Base.metadata.create_all(engine)
            from terms.core.network import Network
            Network.initialize(Session())
        self.compiler = Compiler(Session(), config)
        register_exec_global(Runtime(self.compiler), name='runtime')

    def _parse_buff(self):
        resp = self.compiler.parse(self._buffer)
        self.compiler.session.commit()
        return resp
#        conn = Client((self.config['kb_host'], int(self.config['kb_port'])))
#        conn.send_bytes(self._buffer.encode('ascii'))
#        while True:
#            resp = conn.recv_bytes().decode('ascii')
#            if resp == 'END':
#                conn.close()
#                break
#            resp = self.format_results(resp)
#            print(resp)

    def reset_state(self):
        self._buffer = ''
        self.prompt = '>> '

    def format_results(self, res):
        if isinstance(res, str):
            return res
        resps = [', '.join([k + ': ' + str(v) for k, v in r.items()])
                 for r in res]
        return '; '.join(resps)

    def process_line(self, line):
        self.prompt = '.. '
        resp = self.no_response
        if line:
            self._buffer = '\n'.join((self._buffer, line))
            if self._buffer.endswith(('.', '?')):
                resp = self._parse_buff()
                self.reset_state()
        return resp

    def run(self):
        ic = InteractiveConsole()
        while True:
            line = ic.raw_input(prompt=self.prompt)
            if line in ('quit', 'exit'):
                self.quit()
            resp = self.process_line(line)
            if resp is not self.no_response:
                print(resp)

    def quit(self):
        if HAS_READLINE and self.config['terms_history_file'] and int(self.config['terms_history_length']):
            readline.write_history_file(os.path.expanduser(self.config['terms_history_file']))
        sys.exit('bye')


try:
    import cProfile as profile
except ImportError:
    import profile

def repl():
    config = get_config()
    tr = TermsRepl(config)
    profile.run(tr.run())



# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

import operator

from sqlalchemy import Table, Column, Sequence
from sqlalchemy import ForeignKey, Integer, String, Boolean
from sqlalchemy.orm import relationship, backref, aliased
from sqlalchemy import sql

from terms.core.terms import get_bases
from terms.core.terms import Base, Term
from terms.core.terms import isa
from terms.core.utils import Match




class FactSet(object):
    """
    """

    def __init__(self, name, lexicon, config):
        self.name = name
        self.config = config
        self.session = lexicon.session
        self.lexicon = lexicon

    def get_paths(self, pred):
        '''
        build a path for each testable feature in term.
        Each path is a tuple of strings,
        and corresponds to a node in the primary network.
        '''
        paths = []
        self._recurse_paths(pred, paths, ())
        return paths

    def _recurse_paths(self, pred, paths, path):
        paths.append(path + ('_verb',))
        if not isa(pred, self.lexicon.verb):  # not a verb var
            paths.append(path + ('_neg',))
        for label in sorted(pred.objects):
            o = pred.objects[label].value
            if isa(o, self.lexicon.exist):
                self._recurse_paths(o, paths, path + (label,))
            elif isa(o, self.lexicon.number):
                paths.append(path + (label, '_num'))
            else:
                paths.append(path + (label, '_term'))

    def _get_nclass(self, path):
        ntype = path[-1]
        mapper = Segment.__mapper__
        return mapper.base_mapper.polymorphic_map[ntype].class_

    def add_fact(self, pred):
        print(pred)
        fact = Fact(pred, self.name)
        paths = self.get_paths(pred)
        for path in paths:
            cls = self._get_nclass(path)
            value = cls.resolve(pred, path, self)
            cls(fact, value, path)
        self.session.add(fact)
        self.session.flush()
        return fact

    def add_object_to_fact(self, fact, value, path):
        cls = self._get_nclass(path)
        segment = cls(fact, value, path)
        self.session.add(segment)
        fact.pred.add_object(path[-2], value)

    def query_facts(self, pred, taken_vars, with_factset=True):
        vars = []
        sec_vars = []
        paths = self.get_paths(pred)
        qfacts = self.session.query(Fact)
        if with_factset:
            qfacts = qfacts.filter(Fact.factset==self.name)
        for path in paths:
            cls = self._get_nclass(path)
            value = cls.resolve(pred, path, self)
            if value is not None:
                qfacts = cls.filter_segment(qfacts, value, vars, path)
        vars.sort(key=lambda x: 1 if getattr(x, 'set_condition', False) else 0)
        for var in vars:
            qfacts = var['cls'].filter_segment_first_var(qfacts, var['value'], var['path'], self, taken_vars, sec_vars)
        for var in sec_vars:
            qfacts = var['cls'].filter_segment_sec_var(qfacts, var['path'], var['first'])
        return qfacts

    def query(self, pred):
        taken_vars = {}
        qfacts = self.query_facts(pred, taken_vars)
        matches = []
        for fact in qfacts:
            match = Match(fact.pred, query=pred)
            match.fact = fact
            for name, path in taken_vars.items():
                cls = self._get_nclass(path[0])
                preds = True
                if 'Verb' in name[1:]:
                    preds = False
                value = cls.resolve(fact.pred, path[0], self, preds=preds)
                match[name] = value
            matches.append(match)
        return matches


class Fact(Base):
    __tablename__ = 'facts'

    id = Column(Integer, Sequence('fact_id_seq'), primary_key=True)
    pred_id = Column(Integer, ForeignKey('predicates.id'), index=True)
    pred = relationship('Predicate', backref=backref('facts'),
                         cascade='all',
                         primaryjoin="Predicate.id==Fact.pred_id")
    factset = Column(String(16))

    def __init__(self, pred, name):
        self.pred = pred
        self.factset = name


class Segment(Base):
    __tablename__ = 'segments'

    id = Column(Integer, Sequence('segment_id_seq'), primary_key=True)
    fact_id = Column(Integer, ForeignKey('facts.id'), index=True)
    fact = relationship('Fact',
                         backref='segments',
                         primaryjoin="Fact.id==Segment.fact_id")
    path = Column(String, index=True)

    ntype = Column(String(5))
    __mapper_args__ = {'polymorphic_on': ntype}

    def __init__(self, fact, value, path):
        self.fact = fact
        self.value = value
        self.path = '.'.join(path)

    @classmethod
    def filter_segment(cls, qfact, value, vars, path):
        if getattr(value, 'var', False):
            vars.append({'cls': cls, 'value': value, 'path': path})
        else:
            alias = aliased(cls)
            path_str = '.'.join(path)
            qfact = qfact.join(alias, Fact.id==alias.fact_id).filter(alias.value==value, alias.path==path_str)
        return qfact

    @classmethod
    def resolve(cls, term, path, factset, preds=False):
        '''
        Get the value pointed at by path in w (a word).
        It can be a boolean (for neg nodes),
        a sting (for label nodes),
        a word, or some custom value for custom node types.
        '''
        for segment in path[:-1]:
            term = term.get_object(segment)
        return term

    @classmethod
    def filter_segment_sec_var(cls, qfacts, path, salias):
        alias = aliased(cls)
        path_str = '.'.join(path)
        qfacts = qfacts.join(alias, Fact.id==alias.fact_id).filter(alias.path==path_str, alias.term_id==salias.term_id)
        return qfacts


class NegSegment(Segment):

    __mapper_args__ = {'polymorphic_identity': '_neg'}
    value = Column(Boolean, index=True)

    @classmethod
    def resolve(cls, pred, path, factset, preds=False):
        try:
            for segment in path[:-1]:
                pred = pred.get_object(segment)
            return pred.true
        except AttributeError:
            return None


class TermSegment(Segment):

    __mapper_args__ = {'polymorphic_identity': '_term'}
    term_id = Column(Integer, ForeignKey('terms.id'), index=True)
    value = relationship('Term',
                         primaryjoin="Term.id==TermSegment.term_id")

    @classmethod
    def filter_segment_first_var(cls, qfacts, value, path, factset, taken_vars, sec_vars):
        salias = aliased(cls)
        talias = aliased(Term)
        if value.name in taken_vars:
            sec_vars.append({'cls': cls, 'path': path, 'first': taken_vars[value.name][1]})
            return qfacts
        else:
            taken_vars[value.name] = (path, salias)
        path_str = '.'.join(path)
        if value.bases:
            sbases = [b.id for b in factset.lexicon.get_subterms(value.bases[0])]
            qfacts = qfacts.join(salias, Fact.id==salias.fact_id).filter(salias.path==path_str).join(talias, salias.term_id==talias.id).filter(talias.id.in_(sbases))
        else:
            sbases = [b.id for b in factset.lexicon.get_subterms(value.term_type)]
            qfacts = qfacts.join(salias, Fact.id==salias.fact_id).filter(salias.path==path_str).join(talias, salias.term_id==talias.id).filter(talias.type_id.in_(sbases))
        return qfacts


class NumberSegment(Segment):

    __mapper_args__ = {'polymorphic_identity': '_num'}
    int_value = Column(Integer, index=True)

    binopers = {
        '|': sql.or_,
        '&': sql.and_,
        '=': operator.eq,
        '!=': operator.ne,
        '<': operator.lt,
        '<=': operator.le,
        '>': operator.gt,
        '>=': operator.ge,
        '+': operator.add,
        '-': operator.sub,
        '*': operator.mul,
        '/': operator.truediv,
        '%': operator.mod,
    }
    unopers = {
        '~': sql.not_,
        '-': operator.neg,
    }

    def __init__(self, fact, value, path):
        self.fact = fact
        if getattr(value, 'name', False):
            value = int(value.name)
        self.value = value
        self.path = '.'.join(path)

    @property
    def value(self):
        return self.int_value

    @value.setter
    def value(self, val):
        self.int_value = val

    @classmethod
    def filter_segment(cls, qfact, value, vars, path):
        if getattr(value, 'var', False):
            vars.append({'cls': cls, 'value': value, 'path': path})
        else:
            alias = aliased(cls)
            path_str = '.'.join(path)
            qfact = qfact.join(alias, Fact.id==alias.fact_id).filter(alias.int_value==int(value.name), alias.path==path_str)
        return qfact

    @classmethod
    def filter_segment_first_var(cls, qfacts, value, path, factset, taken_vars, sec_vars):
        alias = aliased(cls)
        taken_vars[value.name] = (path, alias)
        path_str = '.'.join(path)
        qfacts = qfacts.join(alias, Fact.id==alias.fact_id).filter(alias.path==path_str)
        if getattr(value, 'set_condition', False):
            condition = cls.compile_condition(value.set_condition, taken_vars)
            qfacts = qfacts.filter(condition)
        return qfacts

    @classmethod
    def compile_condition(cls, expr, taken_vars):
        if expr.type == 's-vnum':
            return cls.compile_vnum(expr, taken_vars)
        oper = expr.oper
        arg1 = cls.compile_condition(expr.arg1, taken_vars)
        if expr.arg2 is not None:
            arg2 = cls.compile_condition(expr.arg2, taken_vars)
            return cls.binopers[oper](arg1, arg2)
        return cls.unopers[oper](arg1)

    @classmethod
    def compile_vnum(cls, vnum, taken_vars):
        if vnum.var:
            alias = taken_vars[vnum.val][1]
            return getattr(alias, 'int_value')
        return int(vnum.val)


class VerbSegment(Segment):

    __mapper_args__ = {'polymorphic_identity': '_verb'}
    verb_id = Column(Integer, ForeignKey('terms.id'), index=True)
    value = relationship('Term',
                         primaryjoin="Term.id==VerbSegment.verb_id")

    @classmethod
    def resolve(cls, term, path, factset, preds=False):
        for segment in path[:-1]:
            term = term.get_object(segment)
        if term.var or preds:
            return term
        return term.term_type

    @classmethod
    def filter_segment_first_var(cls, qfacts, value, path, factset, taken_vars, sec_vars):
        salias = aliased(cls)
        talias = aliased(Term)
        if value.name in taken_vars:
            sec_vars.append({'cls': cls, 'path': path, 'first': taken_vars[value.name][1]})
            return qfacts
        else:
            taken_vars[value.name] = (path, salias)
#        if value.name == 'Exists1':
#            import pdb;pdb.set_trace()
        if isa(value, factset.lexicon.verb):
            sbases = factset.lexicon.get_subterms(get_bases(value)[0])
        elif isa(value, factset.lexicon.exist):
            sbases = factset.lexicon.get_subterms(value.term_type)
        sbases = [b.id for b in sbases]
        path_str = '.'.join(path)
        qfacts = qfacts.join(salias, Fact.id==salias.fact_id).filter(salias.path==path_str).join(talias, salias.verb_id==talias.id).filter(talias.id.in_(sbases))
        return qfacts

    @classmethod
    def filter_segment_sec_var(cls, qfacts, path, salias):
        alias = aliased(cls)
        path_str = '.'.join(path)
        qfacts = qfacts.join(alias, Fact.id==alias.fact_id).filter(alias.path==path_str, alias.verb_id==salias.verb_id)
        return qfacts



# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.


import os
import os.path
import sys
from configparser import ConfigParser
from optparse import OptionParser


class Match(dict):

    def __init__(self, pred, query=None, prem=None):
        self.pred = pred
        self.paths = []
        self.query = query
        self.prem = prem
        self.building = None
        self.orig_path = ()
        self.fact = None
        self.ancestor = None
        super(Match, self).__init__()

    def copy(self):
        new_match = Match(self.pred)
        for k, v in self.items():
            new_match[k] = v
        new_match.prem = self.prem
        new_match.query = self.query
        new_match.paths = self.paths[:]
        if self.building:
            new_match.building = self.building.copy()
            for l, o in self.items():
                if o is self.building:
                    new_match[l] = new_match.building
                    break
        new_match.orig_path = self.orig_path
        new_match.fact = self.fact
        if self.ancestor:
            new_match.ancestor = self.ancestor.copy()
        return new_match

    def merge(self, m):
        new_match = Match(self.pred)
        tot = {}
        tot.update(self)
        tot.update(m)
        for k, v in tot.items():
            if k in self and k in m and self[k] != m[k]:
                return False
            new_match[k] = v
        return new_match


def merge_submatches(submatches):
    final = []
    while submatches:
        final = submatches.pop()
        if not final:
            return final
        elif not final[0]:
            continue
        break
    while submatches:
        sm = submatches.pop()
        if not sm:
            return sm
        elif not sm[0]:
            continue
        new = []
        for n in sm:
            for m in final:
                nm = m.merge(n)
                if nm:
                    new.append(nm)
        final = new
    return final


def get_config(cmd_line=True):
    config = ConfigParser()
    d = os.path.dirname(sys.modules['terms.core'].__file__)
    fname = os.path.join(d, 'etc', 'terms.cfg')
    config.readfp(open(fname))
    config.read([os.path.expanduser('~/.terms.cfg'),
                 os.path.join('etc', 'terms.cfg')])
    name = 'default'
    if cmd_line:
        parser = OptionParser(usage="usage: %prog [options] [name]")
        parser.add_option("-c", "--config", help="path to config file.")
        opt, args = parser.parse_args()
        name = args[0] if args else name
        if opt.config:
            config.read([opt.config])
    return config[name]



# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

import re

VAR_PAT = r'([A-Z][a-z-]*)([A-Z][a-z-]*)?(\d*)'

varpat = re.compile(VAR_PAT)

SYMBOL_PAT = r'[a-z][a-z-_]*\d*'

sympat = re.compile(SYMBOL_PAT)

NAME_PAT = re.compile(r'^([a-z][a-z-]*[a-z])[1-9]+$')

NUM_PAT = r'(([+]|[-])?([0-9]*)(\.)?([0-9]+)(\.)?)+(j)?((e|E)([0-9]+))?'

numpat = re.compile(NUM_PAT)



# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

class TermsException(Exception):
    pass

class TermRepeated(TermsException):
    pass


class TermNotFound(TermsException):
    pass


class MissingObject(TermsException):
    pass


class Contradiction(TermsException):
    pass


class IllegalLabel(TermsException):
    pass


class WrongLabel(TermsException):
    pass


class Corruption(TermsException):
    pass


class TermsSyntaxError(TermsException):
    pass


class WrongObjectType(TermsException):
    pass

class ImportProblems(TermsException):
    pass




import os
import sys
import logging


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """
    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ''

    def write(self, buf):
        self.logger.log(self.log_level, buf)

    def flush(self):
        for handler in self.logger.handlers:
            handler.flush()


def get_rlogger(config):
    # set up logging
    pname = sys.argv[0]
    logger = logging.getLogger(pname)
    log_file = os.path.abspath(config['logfile'])
    log_dir = os.path.dirname(log_file)
    if not os.path.isfile(log_file):
        if not os.path.isdir(log_dir):
            os.mkdir(log_dir)
        f = open(log_file, 'w')
        f.write('log file for %s\n\n' % pname)
        f.close()
    hdlr = logging.FileHandler(log_file)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    logger.setLevel(getattr(logging, config['loglevel']))
    reader_logger = StreamToLogger(logger)
    return reader_logger



"""Generic linux daemon base class for python 3.x.
from
http://www.jejik.com/files/examples/daemon3x.py"""

import sys
import os
import atexit
import signal


class Daemon:
    """A generic daemon class.

    Usage: subclass the daemon class and override the run() method."""

    def __init__(self, pidfile):
        self.pidfile = pidfile

    def daemonize(self):
        """Deamonize class. UNIX double fork mechanism."""

        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write('fork #1 failed: {0}\n'.format(err))
            sys.exit(1)

        # decouple from parent environment
        os.chdir('/')
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:

                # exit from second parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write('fork #2 failed: {0}\n'.format(err))
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(os.devnull, 'r')
        so = open(os.devnull, 'a+')
        se = open(os.devnull, 'a+')

        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)
        pid = str(os.getpid())
        with open(self.pidfile, 'w+') as f:
            f.write(pid + '\n')

        # signal handling
        signal.signal(signal.SIGTERM, self.cleanup)
        signal.signal(signal.SIGUSR1, self.reload_config)

    def delpid(self):
        os.remove(self.pidfile)

    def start(self):
        """Start the daemon."""

        # Check for a pidfile to see if the daemon already runs
        try:
            with open(self.pidfile, 'r') as pf:

                pid = int(pf.read().strip())
        except IOError:
            pid = None

        if pid:
            message = "pidfile {0} already exist. " + \
                    "Daemon already running?\n"
            sys.stderr.write(message.format(self.pidfile))
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        self.run()

    def cleanup(self, signum, frame):
        """cleanup tasks"""

    def reload_config(self, signum, frame):
        """ reload configuration"""

    def run(self):
        """You should override this method when you subclass Daemon.

        It will be called after the process has been daemonized by
        start() or restart()."""




import json
from terms.core.compiler import Compiler


class _Word(object):
    def __init__(self, id, type=None, bases=(), args=None):
        self.id = id
        self.type = type
        self.bases = bases
        self.args = args or {}

    @classmethod
    def _from_term(cls, term):
        return cls(term.name, term.term_type.name)

    def __call__(self, id, *bases, **kwargs):
        new = _Word(id, type=self)
        if not bases:
            if are(self, verb):
                bases = (exists,)
            if are(self, noun):
                bases = (thing,)
        args = {}
        for base in reversed(bases):
            bases += base.bases
            args.update(getattr(base, 'args', {}))
        new.bases = tuple(set(bases))
        if isa(self, verb):
            new.args = args
            new.args.update(kwargs)
        return new

    def __str__(self):
        return self.id

    __repr__ = __str__

    def __equals__(self, w):
        if self.id == w.id:
            return True
        return False

    def define(self):
        if self.bases:
            sb = ', '.join([b.id for b in self.bases])
            if self.args:
                sa = ', '.join(['%s %s' % (label, a.id) for label, a in self.args.items()])
                s = '%s is %s, %s' % (self.id, sb, sa)
            else:
                s = 'a %s is a %s' % (self.id, sb)
        else:
            s = '%s is a %s' % (self.id, self.type.id)
        return s


def isa(w1, w2):
    if w1.type == w2 or w2 in w1.type.bases:
        return True
    return False


def are(w1, w2):
    if w1 == w2 or w2 in w1.bases:
        return True
    return False


word = _Word('word')
word.type = word
noun = _Word('noun', type=word, bases=(word,))
verb = _Word('verb', type=word, bases=(word,))
number = _Word('number', type=word)
thing = _Word('thing', type=noun, bases=(word,))
exists = _Word('exists', type=verb, bases=(word,), args={'subject': word})


class fact(object):
    def __init__(self, subject, verb_, **kwargs):
        self.subject = subject
        self.type = verb_
        self.bases = ()
        self.args = kwargs

    def __str__(self):
        args = ['%s %s' % (label, str(w)) for label, w in self.args.items()]
        args = ', '.join(args)
        if args:
            return '(%s %s, %s)' % (str(self.type), str(self.subject), args)
        return '(%s %s)' % (str(self.type), str(self.subject))


# >>> import api
# >>> animal = api.noun('animal')
# >>> person = api.noun('person', animal)
# >>> john = person('john')
# >>> anne = person('anne')
# >>> loves = api.verb('loves', who=person)
# >>> f1 = api.fact(john, loves, who=anne)
# >>> wants = api.verb('wants', what=api.exists)
# >>> f2 = api.fact(anne, wants, what=f1)
# >>> str(f2)
# '(wants anne, what (loves john, who anne))'
# >>> john.type.bases
# (animal, thing, word)
# >>> api.isa(john, api.word)
# True
# >>> api.isa(f1, loves)
# True
# >>> Person1 = person('Person1')
# >>> f2 = api.fact(Person1, wants, what=f1)
# >>> str(f2)
# '(wants Person1, what (loves john, who anne))'


class KnowledgeBase(object):

    def __init__(self, session, config):
        self._config = config
        self._sa_session = session
        self._compiler = Compiler(session, config)
        self._network = self._compiler.network
        self._lexicon = self._compiler.lexicon
        self._present = self._network.present
        self._memory = self._network.past

    def tell(self, s):
        return self._compiler.parse(s)

    def tell_word(self, w):
        return self.tell(w.define())

    def tell_fact(self, f):
        return self.tell(str(f))

    def ask_fact(self, f):
        resp = self._network.query(str(f))
        resp = ({label: str(obj) for label, obj in row} for row in resp)
        return resp

    def get_words(self, type):
        term_type = self._lexicon.get_term(type)
        terms = self._lexicon.get_terms(term_type)
        return json.dumps(terms)

    def get_subwords(self, base):
        sterm = self._lexicon.get_term(base)
        terms = self._lexicon.get_subterms(sterm)
        return json.dumps(terms)



# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

from urllib.request import urlopen

import ply.lex as lex
import ply.yacc
from ply.lex import TOKEN


from sqlalchemy.orm.exc import NoResultFound

from terms.core.patterns import SYMBOL_PAT, VAR_PAT, NUM_PAT
from terms.core.network import Network, CondIsa, CondIs, CondCode
from terms.core.terms import isa, Predicate, Import
from terms.core.exceptions import TermsSyntaxError, WrongObjectType, WrongLabel, ImportProblems



class Lexer(object):

    states = (
        ('pycode', 'exclusive'),
        ('set', 'exclusive'),
        ('headers', 'exclusive'),
    )

    tokens = (
        'SYMBOL',
        'NUMBER',
        'COMMA',
        'LPAREN',
        'RPAREN',
        'DOT',
        'QMARK',
        'NOT',
        'IS',
        'TO',
        'A',
        'SEMICOLON',
        'COLON',
        'VAR',
        'IMPLIES',
        'INSTANT_IMPLIES',
        'RM',
        'PYCODE',
        'HEADER',
        'IMPORT',
        'URL',
        'SVAR',
        'SCOLON',
        'SLPAREN',
        'SRPAREN',
        'SNUMBER',
        'SOPER',
        'SPRED',
        'SBAR',
        'SAMP',
        'SNOT',
    )

    reserved = {
        'is': 'IS',
        'to': 'TO',
        'a': 'A',
        'import': 'IMPORT',
    }

    t_NUMBER = NUM_PAT
    t_COMMA = r','
    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_DOT = r'\.'
    t_QMARK = r'\?'
    t_NOT = r'!'
    t_SEMICOLON = r';'
    t_COLON = r':'
    t_VAR = VAR_PAT
    t_INSTANT_IMPLIES = r'-->'
    t_IMPLIES = r'->'
    t_RM = r'_RM_'
    t_URL = r'<[^>]+>'

    t_set_SVAR = VAR_PAT
    t_set_SCOLON = r':'
    t_set_SLPAREN = r'\('
    t_set_SRPAREN = r'\)'
    t_set_SNUMBER = NUM_PAT
    t_set_SOPER = r'[*/+%-]'
    t_set_SPRED = r'[<>=!]=?'
    t_set_SBAR = r'\|'
    t_set_SAMP = r'&'
    t_set_SNOT = r'~'

    @TOKEN(SYMBOL_PAT)
    def t_SYMBOL(self, t):
        # Check for reserved words
        t.type = self.reserved.get(t.value, 'SYMBOL')
        return t

    # Define a rule so we can track line numbers
    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    # A string containing ignored characters (spaces and tabs)
    t_ignore = ' \t'

    def t_begin_pycode(self, t):
        r'<-'
        t.lexer.begin('pycode')

    # Define a rule so we can track line numbers
    def t_pycode_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    def t_pycode_IMPLIES(self, t):
        r'->'
        t.lexer.begin('INITIAL')
        return t

    # Any sequence of characters
    def t_pycode_PYCODE(self, t):
        r'.+'
        return t

    t_pycode_ignore = ''

    def t_begin_set(self, t):
        r'(?<=[^{]){(?=[^{])'
        t.lexer.begin('set')

    def t_set_RCURLY(self, t):
        r' }'
        t.lexer.begin('INITIAL')

    t_set_ignore = ' \t'

    def t_begin_headers(self, t):
        r'{{{'
        t.lexer.begin('headers')

    # Define a rule so we can track line numbers
    def t_headers_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    def t_headers_close(self, t):
        r'}}}'
        t.lexer.begin('INITIAL')

    # Any sequence of characters
    def t_headers_HEADER(self, t):
        r'.+'
        return t

    t_headers_ignore = ''

    # Error handling rule
    def t_set_headers_pycode_INITIAL_error(self, t):
        print("Illegal character '%s' at line %d" % (t.value[0], t.lexer.lineno))
        t.lexer.skip(1)

    # Build the lexer
    def build(self, **kwargs):
        self.lexer = lex.lex(module=self, **kwargs)

    # Test it output
    def test(self, data):
        self.lexer.input(data)
        while True:
            tok = self.lexer.token()
            if not tok:
                break
            print(tok)


class Parser(object):

    precedence = (
        ('left', 'COMMA'),
        ('left', 'LPAREN'),
        ('left', 'SEMICOLON'),
    )

    def __init__(
            self,
            lex_optimize=False,
            yacc_optimize=True,
            yacc_debug=False):

        self.lex = Lexer()

        self.lex.build(
            optimize=lex_optimize)
        self.tokens = self.lex.tokens

        self.parser = ply.yacc.yacc(
            module=self,
            start='module',
            write_tables=False,
            debug=yacc_debug,
            optimize=yacc_optimize)

    def parse(self, text, filename='', debuglevel=0):
        """
            text:
                A string containing the source code

            filename:
                Name of the file being parsed (for meaningful
                error messages)

            debuglevel:
                Debug level to yacc
        """
        self.lex.filename = filename
        # self.lex.reset_lineno()
        return self.parser.parse(text, lexer=self.lex.lexer, debug=debuglevel)

    # BNF

    def p_module(self, p):
        '''module : URL headers constructs
                  | URL constructs
                  | constructs'''
        url, headers, code = None, None, ''
        if len(p) >= 3:
            url = p[1][1:-1]
            if len(p) == 4:
                headers = p[2]
                code = p[3]
            else:
                code = p[2]
        else:
            code = p[1]
        p[0] = AstNode('module', url=url, headers=headers, code=code)

    def p_headers(self, p):
        '''headers : HEADER headers
                   | HEADER'''
        if len(p) == 3:
            p[0] = (p[1],) + p[2]
        else:
            p[0] = (p[1],)

    def p_constructs(self, p):
        '''constructs : construct constructs
                      | construct'''
        if len(p) == 3:
            p[0] = p[2]
            p[0].append(p[1])
        else:
            p[0] = [p[1]]

    def p_construct(self, p):
        '''construct : definition
                     | rule
                     | instant-rule
                     | fact-set
                     | question
                     | removal
                     | import'''
        p[0] = p[1]

    def p_fact_set(self, p):
        '''fact-set : fact-list DOT'''
        p[0] = AstNode('fact-set', facts=p[1])

    def p_definition(self, p):
        '''definition : def DOT'''
        p[0] = AstNode('definition', definition=p[1])

    def p_question(self, p):
        '''question : sentence-list QMARK'''
        p[0] = AstNode('question', facts=p[1])

    def p_removal(self, p):
        '''removal : RM fact-list DOT'''
        p[0] = AstNode('removal', facts=p[2])

    def p_import(self, p):
        '''import : IMPORT URL DOT'''
        p[0] = AstNode('import', url=p[2][1:-1])

    def p_instant_rule(self, p):
        '''instant-rule : sentence-list INSTANT_IMPLIES sentence-list DOT
                        | sentence-list pylines INSTANT_IMPLIES sentence-list DOT'''
        if len(p) == 6:
            pycode = '\n'.join(p[2]).strip()
            p[0] = AstNode('instant-rule', prems=p[1], pycode=pycode, cons=p[4])
        else:
            p[0] = AstNode('instant-rule', prems=p[1], pycode='', cons=p[3])

    def p_rule(self, p):
        '''rule : sentence-list IMPLIES sentence-list DOT
                | sentence-list pylines IMPLIES sentence-list DOT'''
        if len(p) == 6:
            pycode = '\n'.join(p[2]).strip()
            p[0] = AstNode('rule', prems=p[1], pycode=pycode, cons=p[4])
        else:
            p[0] = AstNode('rule', prems=p[1], pycode='', cons=p[3])

    def p_pylines(self, p):
        '''pylines : PYCODE pylines
                   | PYCODE'''
        if len(p) == 3:
            p[0] = (p[1],) + p[2]
        else:
            p[0] = (p[1],)

    def p_fact_list(self, p):
        '''fact-list : fact SEMICOLON fact-list
                     | fact'''
        if len(p) == 4:
            p[0] = (p[1],) + p[3]
        else:
            p[0] = (p[1],)

    def p_sentence_list(self, p):
        '''sentence-list : sentence SEMICOLON sentence-list
                         | sentence'''
        if len(p) == 4:
            p[0] = (p[1],) + p[3]
        else:
            p[0] = (p[1],)

    def p_sentence(self, p):
        '''sentence : def
                    | fact'''
        p[0] = p[1]

    def p_fact(self, p):
        '''fact : prefact
                | var COLON prefact'''
        if len(p) == 4:
            p[3].predvar = p[1]
            p[0] = p[3]
        else:
            p[1].predvar = None
            p[0] = p[1]

    def p_prefact(self, p):
        '''prefact : LPAREN predicate RPAREN
                   | LPAREN NOT predicate RPAREN'''
        if len(p) == 5:
            p[0] = AstNode('fact', predicate=p[3], true=False)
        else:
            p[0] = AstNode('fact', predicate=p[2], true=True)

    def p_predicate(self, p):
        '''predicate : var
                     | verb object
                     | verb object COMMA mods'''
        if len(p) == 2:
            p[0] = AstNode('predicate', verb=p[1], subj=None, mods=())
        elif len(p) == 3:
            p[0] = AstNode('predicate', verb=p[1], subj=p[2], mods=())
        else:
            p[0] = AstNode('predicate', verb=p[1], subj=p[2], mods=p[4])

    def p_verb(self, p):
        '''verb : vterm'''
        p[0] = p[1]

    def p_vterm(self, p):
        '''vterm : term
                 | var'''
        p[0] = p[1]

    def p_vterms(self, p):
        '''vterms : vterm COLON vterms
                  | vterm'''
        if len(p) == 4:
            p[0] = p[3] + (p[1],)
        else:
            p[0] = (p[1],)

    def p_term(self, p):
        '''term : SYMBOL'''
        p[0] = AstNode('term', val=p[1])

    def p_var(self, p):
        '''var : VAR'''
        p[0] = AstNode('var', val=p[1])

    def p_number(self, p):
        '''number : NUMBER'''
        p[0] = AstNode('number', val=p[1])

    def p_mods(self, p):
        '''mods : mod COMMA mods
                | mod'''
        if len(p) == 4:
            p[0] = p[3] + (p[1],)
        else:
            p[0] = (p[1],)

    def p_mod(self, p):
        '''mod : SYMBOL object'''
        p[0] = AstNode('mod', label=p[1], obj=p[2])

    def p_object(self, p):
        '''object : vterm
                  | fact
                  | number
                  | set'''
        p[0] = p[1]

    def p_set(self, p):
        '''set : SVAR SCOLON s-expr'''
        p[0] = AstNode('set', var=p[1], stmnt=p[3])

    def p_snumber(self, p):
        '''s-number : SNUMBER'''
        p[0] = AstNode('s-vnum', val=p[1], var=False)

    def p_svar(self, p):
        '''s-var : SVAR'''
        p[0] = AstNode('s-vnum', val=p[1], var=True)

    def p_svnum(self, p):
        '''s-vnum : s-number
                  | s-var'''
        p[0] = p[1]

    def p_sexpr(self, p):
        '''s-expr : SLPAREN s-expr SRPAREN s-conn SLPAREN s-expr SRPAREN
                  | SNOT SLPAREN s-expr SRPAREN
                  | s-expr s-conn s-expr
                  | s-vnum'''
        if len(p) == 2:
            p[0] = p[1]
        elif len(p) == 4:
            p[0] = AstNode('s-expr', arg1=p[1], oper=p[2], arg2=p[3])
        elif len(p) == 5:
            p[0] = AstNode('s-expr', arg1=p[3], oper=p[1], arg2=None)
        else:
            p[0] = AstNode('s-expr', arg1=p[2], oper=p[4], arg2=p[6])

    def p_sconn(self, p):
        '''s-conn : SOPER
                  | SPRED
                  | SBAR
                  | SAMP
                  | SNOT'''
        p[0] = p[1]

    def p_def(self, p):
        '''def : noun-def
               | name-def
               | verb-def'''
        p[0] = p[1]

    def p_noun_def(self, p):
        '''noun-def : A vterm IS A vterms'''
        p[0] = AstNode('noun-def', name=p[2], bases=p[5])

    def p_terms(self, p):
        '''terms : term COLON terms
                 | term'''
        if len(p) == 4:
            p[0] = p[3] + (p[1],)
        else:
            p[0] = (p[1],)

    def p_name_def(self, p):
        '''name-def : SYMBOL IS A term
                    | vterm IS A vterm'''
        if isinstance(p[1], str):
            p[1] = AstNode('term', val=p[1])
        p[0] = AstNode('name-def', name=p[1], term_type=p[4])

    def p_verb_def(self, p):
        '''verb-def : TO SYMBOL IS TO terms
                    | TO SYMBOL IS TO terms COMMA mod-defs'''
        name = AstNode('term', val=p[2])
        if len(p) == 6:
            p[0] = AstNode('verb-def', name=name, bases=p[5], objs=())
        else:
            p[0] = AstNode('verb-def', name=name, bases=p[5], objs=p[7])

    def p_mod_defs(self, p):
        '''mod-defs : mod-def COMMA mod-defs
                    | mod-def'''
        if len(p) == 4:
            p[0] = p[3] + (p[1],)
        else:
            p[0] = (p[1],)

    def p_mod_def(self, p):
        'mod-def : SYMBOL A term'
        p[0] = AstNode('mod-def', label=p[1], obj_type=p[3])

    def p_error(self, p):
        raise TermsSyntaxError('syntax error: ' + str(p) +
                        ' parsing ' + self.lex.lexer.lexdata)


class AstNode(object):
    def __init__(self, type, **kwargs):
        self.type = type
        for k, v in kwargs.items():
            setattr(self, k, v)


class Compiler(object):

    def __init__(
            self, session, config,
            lex_optimize=False,
            yacc_optimize=True,
            yacc_debug=False):

        self.session = session
        self.config = config
        self.network = Network(session, config)
        self.lexicon = self.network.lexicon

        self.parser = Parser(
            lex_optimize=lex_optimize,
            yacc_optimize=yacc_optimize,
            yacc_debug=yacc_debug)

    def parse(self, s):
        s = '\n'.join([l for l in s.splitlines() if l and not l.startswith('#')])
        module = self.parser.parse(s)
        url = module.url
        headers = module.headers
        known = False
        if url is not None:
            known = True
            try:
                self.session.query(Import).filter_by(url=url).one()
            except NoResultFound:
                known = False
        if not known:
            asts = module.code
            if len(asts) == 1:
                return self.compile(asts[0])
            asts.reverse()
            for ast in asts:
                self.compile(ast)
            if url is not None:  # XXX Save import even if compile throws an exceptin, saving the line it was thrown at?
                headers = '\n'.join(module.headers) if headers is not None else headers
                new = Import(s, url, headers)
                self.session.add(new)
                self.session.commit()
        return 'OK'

    def compile(self, ast):
        if ast.type == 'definition':
            return self.compile_definition(ast.definition)
        elif ast.type == 'rule':
            return self.compile_rule(ast)
        elif ast.type == 'instant-rule':
            return self.compile_instant_rule(ast)
        elif ast.type == 'fact-set':
            return self.compile_factset(ast.facts)
        elif ast.type == 'question':
            return self.compile_question(ast.facts)
        elif ast.type == 'removal':
            return self.compile_removal(ast.facts)
        elif ast.type == 'import':
            return self.compile_import(ast.url)

    def compile_definition(self, definition):
        if definition.type == 'verb-def':
            term = self.compile_verbdef(definition)
        elif definition.type == 'noun-def':
            term = self.compile_noundef(definition)
        elif definition.type == 'name-def':
            term = self.compile_namedef(definition)
        self.session.commit()
        return term

    def compile_verbdef(self, defn):
        bases = [self.lexicon.get_term(t.val) for t in defn.bases]
        objs = {o.label: self.lexicon.get_term(o.obj_type.val)
                for o in defn.objs}
        return self.lexicon.add_subterm(defn.name.val, bases, **objs)

    def compile_noundef(self, defn):
        bases = [self.lexicon.get_term(t.val) for t in defn.bases]
        return self.lexicon.add_subterm(defn.name.val, bases)

    def compile_namedef(self, defn):
        term_type = self.lexicon.get_term(defn.term_type.val)
        return self.lexicon.add_term(defn.name.val, term_type)

    def _prepare_rule(self, rule):
        condcode = None
        if rule.pycode:
            condcode = CondCode(rule.pycode)
        conds, prems = [], []
        for sen in rule.prems:
            if sen.type == 'fact':
                prem = self.compile_fact(sen)
                prems.append(prem)
            else:
                cond = self.compile_conddef(sen)
                conds.append(cond)
        consecs = []
        for sen in rule.cons:
            con = self.compile_fact(sen)
            consecs.append(con)
        return prems, conds, condcode, consecs

    def compile_rule(self, rule):
        args = self._prepare_rule(rule)
        self.network.add_rule(*args)
        self.session.commit()
        return 'OK'

    def compile_instant_rule(self, rule_ast):
        args = self._prepare_rule(rule_ast)
        rule = self.network.add_rule(*args)
        # remove premnodes & nodes that have no other rules
        for prem in rule.prems:
            if len(prem.node.prems) == 1:
                node = prem.node
                while len(node.parent.children) == 1:
                    node = node.parent
                self.session.delete(node)
        self.session.delete(rule)
        return 'OK'

    def compile_fact(self, fact):
        true = fact.true
        verb = self.compile_vterm(fact.predicate.verb)
        if fact.predicate.subj is None:
            return verb
        subj = self.compile_obj(fact.predicate.subj)
        mods = self.compile_mods(verb, fact.predicate.mods)
        mods['subj'] = subj
        redundant_var = None
        if fact.predvar:
            redundant_var = self.lexicon.make_var(fact.predvar.val)
        return Predicate(true, verb, redundant_var_=redundant_var, **mods)

    def compile_vterm(self, vterm):
        if vterm.type == 'var':
            return self.lexicon.make_var(vterm.val)
        return self.lexicon.get_term(vterm.val)

    def compile_obj(self, obj):
        if obj.type == 'var':
            return self.lexicon.make_var(obj.val)
        elif obj.type == 'term':
            return self.lexicon.get_term(obj.val)
        elif obj.type == 'fact':
            return self.compile_fact(obj)
        elif obj.type == 'number':
            return self.lexicon.make_term(obj.val, self.lexicon.number)
        elif obj.type == 'set':
            return self.compile_set(obj)

    def compile_set(self, ast):
        var = self.lexicon.make_var(ast.var)
        var.set_condition = ast.stmnt
        return var

    def compile_mods(self, verb, ast):
        mods = {}
        for mod in ast:
            label = mod.label
            try:
                otype = tuple(filter(lambda x: x.label == label, verb.object_types))[0]
            except IndexError:
                raise WrongLabel('Error: label %s is unknown for verb %s (in a rule)' % (label, verb.name))
            obj = self.compile_obj(mod.obj)
            if not isa(obj, otype.obj_type):
                raise WrongObjectType('Error: word %s for label %s is not the correct type: '
                                       'it is a %s and should be a %s' %
                                      (obj.name, label, obj.term_type.name, otype.obj_type.name))
            mods[label] = obj
        return mods

    def compile_conddef(self, sen):
        if sen.type == 'name-def':
            name = self.compile_vterm(sen.name)
            term_type = self.compile_vterm(sen.term_type)
            return CondIsa(name, term_type)
        else:
            name = self.compile_vterm(sen.name)
            base = self.compile_vterm(sen.bases[0])
            return CondIs(name, base)

    def compile_factset(self, facts):
        for f in facts:
            pred = self.compile_fact(f)
            self.network.add_fact(pred)
            self.session.commit()
        return 'OK'

    def compile_question(self, sentences):
        matches = []
        if sentences:
            facts, defs = [], []
            for s in sentences:
                if s.type == 'fact':
                    facts.append(s)
                else:
                    defs.append(s)
            q = [self.compile_fact(f) for f in facts]
            matches = self.network.query(*q)
            for defn in defs:
                if defn.type == 'noun-def':
                    if defn.name.type == 'var':
                        terms = self.lexicon.get_terms
                        #  XXX unfinished
                elif defn.type == 'name-def':
                    term = self.compile_namedef(defn)

        if not matches:
            matches = 'false'
        elif not matches[0]:
            matches = 'true'
        return matches

    def compile_removal(self, facts):
        for f in facts:
            pred = self.compile_fact(f)
            self.network.del_fact(pred)
        self.session.commit()
        return 'OK'

    def compile_import(self, url):
        try:
            self.session.query(Import).filter_by(url=url).one()
        except NoResultFound:
            if url.startswith('file://'):
                path = url[7:]
                try:
                    f = open(path, 'r')
                except Exception as e:
                    raise ImportProblems('Problems opening the file: ' + str(e))
                code = f.read()
                f.close()
            elif url.startswith('http'):
                try:
                    resp = urlopen(url)
                except Exception as e:
                    raise ImportProblems('Problems loading the file: ' + str(e))
                code = resp.read().decode('utf8')
                resp.close()
            else:
                raise ImportProblems('Unknown protocol for <%s>' % url)
            self.parse(code)
        return 'OK'


class Runtime(object):

    def __init__(self, compiler):
        self.compiler = compiler

    def count(self, sen):
        resp = self.compiler.parse(sen + '?')
        if resp == 'false':
            return 0
        elif resp == 'true':
            return 1
        return len(resp)



import os
import sys

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from terms.core.utils import get_config
from terms.core.network import Network
from terms.core.terms import Base


def init_terms():
    config = get_config()
    address = '%s/%s' % (config['dbms'], config['dbname'])
    engine = create_engine(address)
    Base.metadata.create_all(engine)
    Session = sessionmaker(bind=engine)
    session = Session()
    Network.initialize(session)
    session.commit()
    session.close()
    sys.exit('Created knowledge store %s' % config['dbname'])



# -*- coding: utf-8 -*-
# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.


import distribute_setup
distribute_setup.use_setuptools()
from setuptools import setup, find_packages

VERSION = '0.1.0b8'

setup(
    name = 'Terms',
    version = VERSION,
    author = 'Enrique Pérez Arnaud',
    author_email = 'enriquepablo@gmail.com',
    url = 'http://terms.readthedocs.org/en/latest/',
    license = 'GNU GENERAL PUBLIC LICENSE Version 3',
    description = 'A rule production system',
    platform = 'UNIX',
    classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Operating System :: POSIX",
          "Programming Language :: Python :: 3",
          "Topic :: Scientific/Engineering",

    ],
    long_description = (open('README.rst').read() +
                        '\n' + open('INSTALL.rst').read()) +
                        '\n' + open('SUPPORT.rst').read(),

    packages = find_packages(),
    namespace_packages = ['terms'],
    test_suite = 'nose.collector',
    include_package_data = True,

    entry_points = {
        'console_scripts': [
            'terms = terms.core.scripts.repl:repl',
            'initterms = terms.core.scripts.initterms:init_terms',
            'kbdaemon = terms.core.scripts.kbdaemon:main',
        ],
    },
    tests_require = [
        'Nose',
        'coverage',
    ],
    extras_require = {
        'PG': ['psycopg2 == 2.5.1',],
        },
    install_requires = [
        'distribute',
        'sqlalchemy >= 0.8.2',
        'ply == 3.4',
    ],
)



import sys
import os
import time
import signal

from terms.core.kb import KnowledgeBase
from terms.core.utils import get_config


def start(config):
    config['logfile'] = os.path.abspath(config['logfile'])
    kb = KnowledgeBase(config)
    kb.start()


def stop(config):
    # Get the pid from the pidfile
    pidfile = os.path.abspath(config['pidfile'])
    try:
        with open(pidfile, 'r') as pf:
            pid = int(pf.read().strip())
    except IOError:
        pid = None

    if not pid:
        message = "pidfile {0} does not exist. " + \
                "Daemon not running?\n"
        sys.stderr.write(message.format(pidfile))
        return  # not an error in a restart

    # Try killing the daemon process
    try:
        os.kill(pid, signal.SIGTERM)
    except OSError as err:
        e = str(err.args)
        if e.find("No such process") > 0:
            if os.path.exists(pidfile):
                os.remove(pidfile)
            else:
                print (str(err.args))
    sys.exit(1)


def main():
    config = get_config(cmd_line=False)
    if len(sys.argv) == 2:
        if 'start' == sys.argv[1]:
            start(config)
        elif 'stop' == sys.argv[1]:
            stop(config)
        elif 'restart' == sys.argv[1]:
            stop(config)
            start(config)
        else:
            print("Unknown command")
            sys.exit(2)
        sys.exit(0)
    else:
        print("usage: %s start|stop|restart" % sys.argv[0])
        sys.exit(2)



# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

import sys
import os
from configparser import ConfigParser
import nose

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from terms.core.terms import Base
from terms.core.network import Network
from terms.core.compiler import KnowledgeBase


CONFIG = '''
[test]
dbms = postgresql://terms:terms@localhost
dbname = test
#dbms = sqlite://
#dbname = :memory:
time = normal
import =
instant_duration = 0
'''


def test_terms():  # test generator
    # read contents of tests/
    # feed each test to run_npl
    d = os.path.dirname(sys.modules['terms.core'].__file__)
    d = os.path.join(d, 'tests')
    files = os.listdir(d)
    config = ConfigParser()
    config.read_string(CONFIG)
    config = config['test']
    for f in files:
        if f.endswith('.test'):
            address = '%s/%s' % (config['dbms'], config['dbname'])
            engine = create_engine(address)
            Session = sessionmaker(bind=engine)
            session = Session()
            Base.metadata.create_all(engine)
            Network.initialize(session)
            kb = KnowledgeBase(session, config,
                               lex_optimize=False,
                               yacc_optimize=False,
                               yacc_debug=True)
            yield run_terms, kb, os.path.join(d, f)
            kb.session.close()
            Base.metadata.drop_all(engine)


def run_terms(kb, fname):
    # open file, read lines
    # tell assertions
    # compare return of questions with provided output
    with open(fname) as f:
        resp = kb.no_response
        for sen in f:
            sen = sen.rstrip()
            if resp is not kb.no_response:
                sen = sen.strip('.')
                pmsg = 'returned "%s" is not "%s" at line %d for query: %s'
                msg = pmsg % (resp, sen,
                              kb.parser.lex.lexer.lineno,
                              kb.parser.lex.lexer.lexdata)
                nose.tools.assert_equals(sen, resp, msg=msg)
                resp = kb.no_response
            elif sen and not sen.startswith('#'):
                resp = kb.process_line(sen)



import os
import sys
import time
import json
import multiprocessing as mp
from multiprocessing import Process, JoinableQueue, Lock
from multiprocessing.connection import Listener
from threading import Thread

from sqlalchemy.orm.exc import NoResultFound

from terms.core import register_exec_global
from terms.core.terms import Term, Predicate, isa
from terms.core.terms import ExecGlobal, load_exec_globals
from terms.core.compiler import Compiler, Runtime
from terms.core.sa import get_sasession
from terms.core.daemon import Daemon
from terms.core.logger import get_rlogger

from terms.core.exceptions import TermNotFound, TermsSyntaxError, WrongLabel, IllegalLabel, WrongObjectType, ImportProblems


class TermsJSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if type(obj) in (Term, Predicate):
            return str(obj)
        else:
            return super(TermsJSONEncoder, self).default(obj)


class Teller(Process):

    def __init__(self, config, session_factory, teller_queue, *args, **kwargs):
        super(Teller, self).__init__(*args, **kwargs)
        self.config = config
        self.session_factory = session_factory
        self.teller_queue = teller_queue
        self.compiler = None

    def run(self):
        for client in iter(self.teller_queue.get, None):
            totell = []
            for msg in iter(client.recv_bytes, b'FINISH-TERMS'):
                totell.append(msg.decode('utf8'))
            totell = '\n'.join(totell)
            session = self.session_factory()
            self.compiler = Compiler(session, self.config)
            register_exec_global(Runtime(self.compiler), name='runtime')
            load_exec_globals(session)
            if totell.startswith('lexicon:'):
                resp = self._from_lexicon(totell)
            elif totell.startswith('compiler:exec_globals:'):
                resp = self._add_execglobal(totell)
            else:
                self.compiler.network.pipe = client
                try:
                    resp = self.compiler.parse(totell)
                except TermNotFound as e:
                    session.rollback()
                    resp = 'Unknown word: ' + e.args[0]
                except TermsSyntaxError as e:
                    session.rollback()
                    resp = 'Terms syntax error: ' + e.args[0]
                except WrongLabel as e:
                    session.rollback()
                    resp = e.args[0]
                except IllegalLabel as e:
                    session.rollback()
                    resp = 'Error: labels cannot contain underscores: %s' % e.args[0]
                except WrongObjectType as e:
                    session.rollback()
                    resp = e.args[0]
                except ImportProblems as e:
                    session.rollback()
                    resp = e.args[0]
                self.compiler.network.pipe = None
                resp = json.dumps(resp, cls=TermsJSONEncoder)
            try:
                client.send_bytes(str(resp).encode('utf8'))
            except BrokenPipeError:
                pass
            else:
                client.send_bytes(b'END')
            client.close()
            session.close()  # XXX needed?
            self.compliler = None
            self.teller_queue.task_done()  # abyss
        self.teller_queue.task_done()
        self.teller_queue.close()

    def _from_lexicon(self, totell):
        q = totell.split(':')
        ttype = self.compiler.lexicon.get_term(q[2])
        if q[1] == 'get-words':
            resp = self.compiler.lexicon.get_terms(ttype)
        elif q[1] == 'get-subwords':
            resp = self.compiler.lexicon.get_subterms(ttype)
        elif q[1] == 'get-verb':
            resp = []
            for ot in ttype.object_types:
                isverb = isa(ot.obj_type, self.compiler.lexicon.verb)
                resp.append([ot.label, ot.obj_type.name, isverb])
        return json.dumps(resp, cls=TermsJSONEncoder)

    def _add_execglobal(self, totell):
# XXX put it in terms.core.exec_globals, in all processes
        egs = totell[22:]
        eg = ExecGlobal(egs)
        session = self.session_factory()
        session.add(eg)
        session.commit()
        session.close()


class KnowledgeBase(Daemon):

    def __init__(self, config):
        self.config = config
        self.pidfile = os.path.abspath(config['pidfile'])
        self.time_lock = Lock()
        self.teller_queue = JoinableQueue()
        self.session_factory = get_sasession(self.config)
        session = self.session_factory()

    def run(self):
        reader_logger = get_rlogger(self.config)
        sys.stdout = reader_logger
        sys.stderr = reader_logger

        if int(self.config['instant_duration']):
            self.clock = Ticker(self.config, self.session_factory(),
                                self.time_lock, self.teller_queue)
            self.clock.start()

        host = self.config['kb_host']
        port = int(self.config['kb_port'])
        nproc = int(self.config['teller_processes'])
        for n in range(nproc):
            teller = Teller(self.config, self.session_factory, self.teller_queue)
            teller.daemon = True
            teller.start()
        self.socket = Listener((host, port))
        while True:
            try:
                client = self.socket.accept()
            except InterruptedError:
                return
            self.time_lock.acquire()
            self.teller_queue.put(client)
            self.time_lock.release()

    def cleanup(self, signum, frame):
        """cleanup tasks"""
        nproc = int(self.config['teller_processes'])
        for n in range(nproc):
            self.teller_queue.put(None)
        self.teller_queue.close()
        try:
            self.clock.ticking = False
        except AttributeError:
            pass
        self.teller_queue.join()
        try:
            self.clock.join()
        except AttributeError:
            pass
        print('bye from {n}, received signal {p}'.format(n=mp.current_process().name, p=str(signum)))


class Ticker(Thread):

    def __init__(self, config, session, lock, queue, *args, **kwargs):
        super(Ticker, self).__init__(*args, **kwargs)
        self.config = config
        self.session = session
        self.compiler = Compiler(session, config)
        #register_exec_global(self.compiler, name='compiler')
        self.time_lock = lock
        self.teller_queue = queue
        self.ticking = True

    def run(self):
        while self.ticking:
            self.time_lock.acquire()
            self.teller_queue.join()
            pred = Predicate(True, self.compiler.lexicon.vtime,
                             subj=self.compiler.lexicon.now_term)
            try:
                fact = self.compiler.network.present.query_facts(pred,
                                                                 []).one()
            except NoResultFound:
                pass
            else:
                self.session.delete(fact)
                self.session.commit()
            self.compiler.network.passtime()
            pred = Predicate(True, self.compiler.lexicon.vtime,
                             subj=self.compiler.lexicon.now_term)
            self.compiler.network.add_fact(pred)
            self.session.commit()
            self.time_lock.release()
            if self.ticking:
                time.sleep(float(self.config['instant_duration']))
            else:
                self.session.close()



# Copyright (c) 2007-2012 by Enrique Pérez Arnaud <enriquepablo@gmail.com>
#
# This file is part of the terms project.
# https://github.com/enriquepablo/terms
#
# The terms project is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The terms project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with any part of the terms project.
# If not, see <http://www.gnu.org/licenses/>.

import datetime

from sqlalchemy import Table, Column, Sequence, Index, DateTime
from sqlalchemy import ForeignKey, Integer, String, Boolean, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, backref
from sqlalchemy.orm.collections import attribute_mapped_collection

from terms.core.exceptions import WrongLabel


class Base(object):
    pass

Base = declarative_base(cls=Base)


term_to_base = Table('term_to_base', Base.metadata,
    Column('term_id', Integer, ForeignKey('terms.id')),
    Column('base_id', Integer, ForeignKey('terms.id'))
)

term_to_objtype = Table('term_to_objtype', Base.metadata,
    Column('term_id', Integer, ForeignKey('terms.id')),
    Column('objtype_id', Integer, ForeignKey('objecttypes.id'))
)


class Term(Base):
    __tablename__ = 'terms'

    id = Column(Integer, Sequence('term_id_seq'), primary_key=True)
    name = Column(String)
    type_id = Column(Integer, ForeignKey('terms.id'))
    term_type = relationship('Term', remote_side=[id],
                         primaryjoin="Term.id==Term.type_id",
                         post_update=True)
    bases = relationship('Term', backref='subwords',
                         secondary=term_to_base,
                         primaryjoin=id == term_to_base.c.term_id,
                         secondaryjoin=id == term_to_base.c.base_id)
    object_types = relationship('ObjectType',
                         secondary=term_to_objtype,
                         primaryjoin=id == term_to_objtype.c.term_id,
                         secondaryjoin='ObjectType.id==term_to_objtype.c.objtype_id')
    equals = ()
    var = Column(Boolean)
    number = Column(Boolean, default=False)

    rule_id = Column(Integer, ForeignKey('rules.id'))
    rule = relationship('Rule', backref=backref('vconsecuences', cascade='all'),
                         primaryjoin="Rule.id==Term.rule_id")

    term_name_index = Index('term_name_index', 'name')
    term_type_index = Index('term_type_index', 'type_id')

    # Avoid AttributeErrors
    objects = {}

    def __init__(self, name,
                       ttype=None,
                       bases=None,
                       objs=None,
                       var=False,
                       _bootstrap=False):
        self.name = name
        self.var = var
        self._sup_cache = None
        self._sub_cache = None
        if not _bootstrap:
            self.term_type = ttype or bases[0].term_type
        used = []
        if objs is None:
            objs = {}
        for label, otype in objs.items():
            objtype = ObjectType(label, otype)
            self.object_types.append(objtype)
            used.append(label)
        if bases:
            for base in bases:
                self.bases.append(base)
                for objtype in base.object_types:
                    if objtype.label not in used:
                        self.object_types.append(objtype)
                        used.append(objtype.label)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<Term: %s>' % str(self)

    def copy(self):
        #  immutable
        return self

    def purge_cache(self, lexicon):
        for sub in lexicon.get_subterms(self):
            sub.purge_sup()
        for sup in get_bases(self):
            sup.purge_sub(lexicon)

    def purge_sup(self):
        for sup in get_bases(self):
            sup.purge_sup()
        self._sup_cache = None

    def purge_sub(self, lexicon):
        for sub in lexicon.get_subterms(self):
            if sub is self:
                continue
            sub.purge_sub(lexicon)
        self._sub_cache = None


class ObjectType(Base):
    ''' '''
    __tablename__ = 'objecttypes'

    id = Column(Integer, Sequence('objecttypes_id_seq'), primary_key=True)
    label = Column(String(100))
    obj_type_id = Column(Integer, ForeignKey('terms.id'))
    obj_type = relationship(Term, primaryjoin='Term.id==ObjectType.obj_type_id')

    def __init__(self, label, obj_type):
        self.label = label
        self.obj_type = obj_type


class Predicate(Base):
    '''
    Predicates, used for interchange and
    persisted as consecuences.
    '''
    __tablename__ = 'predicates'

    id = Column(Integer, Sequence('predicate_id_seq'), primary_key=True)
    true = Column(Boolean)
    type_id = Column(Integer, ForeignKey('terms.id'))
    term_type = relationship('Term', primaryjoin="Term.id==Predicate.type_id", lazy='joined')
    rule_id = Column(Integer, ForeignKey('rules.id'))
    rule = relationship('Rule', backref=backref('consecuences', cascade='all'),
                         primaryjoin="Rule.id==Predicate.rule_id")

    # to avoid AttributeErrors when used as a term
    bases = ()
    name = ''
    var = False

    def __init__(self, true, verb_, redundant_var_=None, **objs):
        '''
        verb is a string.
        args is a dict with strings (labels) to ConObjects
        '''
        self.true = true
        self.term_type = verb_
        labels = [ot.label for ot in verb_.object_types]
        for label, o in objs.items():
            if label not in labels:
                raise WrongLabel('Error: label "%s" is wrong for verb "%s"' % (label, verb_.name))
            self.add_object(label, o)
        self.redundant_var = redundant_var_

    def __str__(self):
        p = not self.true and '!' or ''
        p += str(self.term_type)
        p = ['%s %s' % (p, str(self.get_object('subj')))]
        for label in sorted(self.objects):
            if label != 'subj':
                p.append('%s %s' % (label, str(self.get_object(label))))
        return '(%s)' % ', '.join(p)

    def __repr__(self):
        return '<Predicate: %s>' % str(self)

    def add_object(self, label, obj):
        if isinstance(obj, Predicate):
            self.objects[label] = PObject(label, obj)
        else:
            self.objects[label] = TObject(label, obj)

    def get_object(self, label):
        return self.objects[label].value

    def substitute(self, match):
        if self.term_type.var:
            new = Predicate(self.true, match[self.term_type.name])
        else:
            new = Predicate(self.true, self.term_type)
        for o in self.objects.values():
            obj = o.copy()
            if isinstance(o.value, Predicate):
                obj.value = o.value.substitute(match)
            elif o.value.var:
                value = match[o.value.name]
                if isinstance(o, TObject) and isinstance(value, Predicate):
                    obj = PObject(o.label, value)
                else:
                    obj.value = value
                    obj.value.var = False
            new.objects[obj.label] = obj
        return new

    def copy(self):
        new = Predicate(self.true, self.term_type)
        for o in self.objects.values():
            if o is not None:
                new.objects[o.label] = o.copy()
        return new

    def get_vars(self, vars=None):
        if vars is None:
            vars = []
        for o in self.objects:
            if o.value.var:
                vars.append(o.value)
            elif isinstance(o.value, Predicate):
                o.value.get_vars(vars)
        return vars


class Object(Base):
    '''
    objects for Predicates
    '''
    __tablename__ = 'objects'

    id = Column(Integer, Sequence('object_id_seq'), primary_key=True)
    parent_id = Column(Integer, ForeignKey('predicates.id'))
    parent = relationship('Predicate',
                          backref=backref('objects',
                                          collection_class=attribute_mapped_collection('label'),
                                          lazy='joined',
                                          cascade='all,delete-orphan'),
                          primaryjoin="Predicate.id==Object.parent_id")
    label = Column(String(100))

    otype = Column(Integer)
    __mapper_args__ = {'polymorphic_on': otype}

    def __init__(self, label, term):
        self.label = label
        self.value = term

    def copy(self):
        cls = type(self)
        nval = self.value and self.value.copy() or self.value
        return cls(self.label, nval)


class TObject(Object):
    '''
    '''
    __mapper_args__ = {'polymorphic_identity': 0}
    term_id = Column(Integer, ForeignKey('terms.id'))
    value = relationship('Term', primaryjoin="Term.id==TObject.term_id", lazy='joined')


class PObject(Object):
    '''
    '''
    __mapper_args__ = {'polymorphic_identity': 1}
    pred_id = Column(Integer, ForeignKey('predicates.id'))
    value = relationship('Predicate', cascade='all', lazy='joined',
                         primaryjoin="Predicate.id==PObject.pred_id")


def isa(t1, t2):
    try:
        ttype = t1.term_type
    except AttributeError:
        return False
    return are(ttype, t2)


def are(t1, t2):
    if t1 == t2:
        return True
    try:
        equals = get_equals(t1, search=t2)
        for eq in equals:
            get_bases(eq, search=t2)
    except SearchFound:
        return True
    return False

def eq(t1, t2):
    if t1 == t2:
        return True
    try:
        get_equals(t1, search=t2)
    except SearchFound:
        return True
    return False

def get_bases(term, search=None):
    cache = getattr(term, '_sup_cache', None)
    if cache is not None:
        if search and search in cache:
            raise SearchFound(search)
        return cache
    bases = _get_desc(term, 'bases', search=search)
    term._sup_cache = bases
    return bases

def get_equals(term, search=None):
    return (term,) + _get_desc(term, 'equals', search=search)

class SearchFound(Exception): pass

def _get_desc(term, desc, search=None, bset=None):
    if not bset:
        bset = set()
    bases = getattr(term, desc, None)
    if bases is None:
        return ()
    for base in bases:
        if search and search == base:
            raise SearchFound(base)
        bset.add(base)
        _get_desc(base, desc, search=search, bset=bset)
        if desc != 'equals':
            for eq in base.equals:
                bset.add(eq)
    return tuple(bset)


class Time(Base):
    '''
    '''
    __tablename__ = 'time'
    id = Column(Integer, default=0, primary_key=True)
    now = Column(Integer, default=0)


class Import(Base):
    '''
    '''
    __tablename__ = 'imports'
    id = Column(Integer, Sequence('import_id_seq'), primary_key=True)
    url = Column(String())
    headers = Column(Text())
    code = Column(Text())
    timestamp = Column(DateTime, default=datetime.datetime.utcnow)

    def __init__(self, code, url, headers):
        self.code = code
        self.url = url
        self.headers = headers


class ExecGlobal(Base):
    '''
    '''
    __tablename__ = 'execglobals'
    id = Column(Integer, Sequence('execglobal_id_seq'), primary_key=True)
    code = Column(Text())

    def __init__(self, code):
        self.code = code


def load_exec_globals(session):
    egs = session.query(ExecGlobal).all()
    from terms.core import localdata
    for eg in egs:
        exec(eg.code, localdata.exec_globals)
